# coding:utf-8
# todo 文件说明写成 README.md
"""
此文件用以处理文件正则，需要指定一个正则规则文件
文件正则文件各式如下

不可命中规则1 不可命中规则2: 命中规则1 命中规则2

规则之间由空格分割，不可命中可以有也可以没有，不可命中和可命中规则均可有多个
在 分号: 前的规则有一个被命中视为与该正则规则不符，
分号: 后的规则有一个被命中，视为符合

规则实例文件见example/process_pattern_regex.exp
"""
from __future__ import absolute_import
from __future__ import print_function
from __future__ import unicode_literals

from longling.base import *
from longling.lib.process_pattern.pattener import line_init_patterns, \
    init_patterns
from longling.lib.process_pattern.process_pattern_base import \
    ProcessPatternEncodedError


def regex(ps, line):
    if not ps:
        return False
    xs = []
    for p in ps:
        x = p.search(line)
        xs.append(x)
    for x in xs:
        if x is None:
            return False
    return True


class PatternHitter(object):
    """
    模式识别器

    Examples
    --------
    >>> p = PatternHitter(":世界 和平", 1)
    >>> print(p.hit_num("世界 和平"))
    1
    >>> print(p.is_in(u"世界和平"))
    True
    >>> p = PatternHitter(":世界 和平", 0)
    >>> print(p.hit_num("世界 和平"))
    1
    >>> print(p.is_in(u"世界和平"))
    False
    """

    def __init__(self, location, mode=0):
        """

        Parameters
        ----------
        location
        mode
        """
        if ':' in location:
            self.pps = dict()
            line_init_patterns(location, mode, self.pps)
        else:
            self.pps = init_patterns(location, mode)

    def is_in(self, line):
        """
        检查输入中是否命中规则

        Parameters
        ----------
        line: string_type

        Returns
        -------
        bool
            命中返回True
        """
        if isinstance(line, str):
            line = unistr(line)
        if not isinstance(line, string_types):
            raise Exception("输入必须是str或unicode")
        for name, (ps1, ps2) in self.pps.items():
            t1 = regex(ps1, line)
            t2 = regex(ps2, line)
            if (len(ps1) == 0 or t1 is not True) and t2 is True:
                # 不满足1的同时满足2才行
                # print line.encode('utf-8'),name.encode('utf-8')
                return True
        return False

    def hit_num(self, line):
        """
        计算输入命中规则的次数

        Parameters
        ----------
        line: string_type

        Returns
        -------
        int
            命中次数
        """
        if isinstance(line, str):
            line = unistr(line)
        if not isinstance(line, string_types):
            raise ProcessPatternEncodedError("输入必须是str或unicode")
        t = 0
        for name, (ps1, ps2) in self.pps.items():
            t1 = regex(ps1, line)
            t2 = regex(ps2, line)
            if (len(ps1) == 0 or t1 is not True) and t2 is True:
                # 不满足1的同时满足2才行
                t += 1
        return t

    def hit_group(self, line):
        hit_group = []
        for name, (ps1, ps2) in self.pps.items():
            t1 = regex(ps1, line)
            t2 = regex(ps2, line)
            if (len(ps1) == 0 or t1 is not True) and t2 is True:
                # 不满足1的同时满足2才行
                hit_group.append(name)
        return hit_group

        # def hit_num(self, line):
        #     return len(self.hit_group(line))


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# coding: utf-8
# create by tongshiwei on 2018/8/16
"""
此模块用来针对result.json中的数据进行数据分析
开发测试中，非稳定版本
"""
import math
import re
from collections import defaultdict

import matplotlib.pyplot as plt

__all__ = ["ResultAnalyser"]


class ResultAnalyser(object):
    """
    Examples
    --------
    >>> result_analyser = ResultAnalyser()
    >>> dict_data = {'key1': {'key2': [1, 2, 3], 'key3': [1, 2]}}
    >>> result_analyser.add_record(dict_data)
    >>> 'key1_key2' in result_analyser.records
    True
    >>> 'key1_key3' in result_analyser
    True
    """

    def __init__(self):
        self.records = defaultdict(list)

    def add_record(self, dict_data, prefix=''):
        """
        添加字典数据并解析，如果字典的value也是字典的话，会深入解析，
        键名会进行拼接操作，如: {'key1': {'key2': value}}
        会返回 (key1_key2, value)

        Parameters
        ----------
        dict_data: dict
            字典格式数据
        prefix: str
        """
        for key, value in dict_data.items():
            if isinstance(value, dict):
                self.add_record(value, prefix=prefix + '_' + str(
                    key) if prefix else str(key))
            else:
                self.records[
                    prefix + '_' + str(key) if prefix else str(key)].append(
                    value)

    def __iter__(self):
        return iter(self.records)

    def keys(self):
        return self.records.keys()

    def items(self):
        return self.records.items()

    def values(self):
        return self.records.values()

    def select(self, select='iteration'):
        pattern = re.compile(select)
        key_value = []
        for key, value in self.items():
            if pattern.match(key):
                key_value.append((key, value))
        return key_value

    def selects(self, selects=('accuracy|prf_avg_*', 'prf_\d+_.*',)):
        patterns = [re.compile(select) for select in selects]
        data = [[] for _ in patterns]
        for key, value in self:
            for i, pattern in enumerate(patterns):
                if pattern.match(key):
                    data[i].append((key, value))
        return data

    def visual_selects(self, x_select='iteration',
                       y_selects=('accuracy|prf_avg_.*', 'prf_\d+_f1')):
        """
        按 x，y 格式选取数据，x_select 和 y_selects 都是正则匹配式，
        采用 re.match 方式匹配

        Parameters
        ----------
        x_select: str
        y_selects: tuple[str] or list[str]
            path select

        Returns
        -------
        x_data: tuple(str, list)
            x_key: str
                代表 x 的名称
            x: list
                代表数据
        y_datas: list[list[tuple(str, list)]]
            每一个元素都是一个列表，列表中的每一个元素是一个元组，
            元组包含两个元素:
                第一个元素，字符串类型，代表数据名称(path_key形式)
                第二个元素，列表，代表数据
        """
        x_pattern = re.compile(x_select)
        y_patterns = [re.compile(y_select) for y_select in y_selects]

        x = None
        x_key = None
        ys = [[] for _ in y_patterns]
        for key, value in self.items():
            if x_pattern.match(key):
                assert x is None, \
                    "x has been set, duplicated x, " \
                    "%s [stored] vs %s [store]" % (
                        x_key, key
                    )
                x_key = key
                x = value
            else:
                for i, y_pattern in enumerate(y_patterns):
                    if y_pattern.match(key):
                        ys[i].append((key, value))

        return (x_key, x), ys

    def visual_select(self, x_select='iteration',
                      y_select='accuracy|prf_avg_.*'):
        """
        按 x，y 格式选取数据，x_select 和 y_select 都是正则匹配式，
        采用 re.match 方式匹配

        Parameters
        ----------
        x_select: str
        y_select: str
            path select

        Returns
        -------
        x_data: tuple(str, list)
            x_key: str
                代表 x 的名称
            x: list
                代表数据
        y_datas: list[tuple(str, list)]
            每一个元素都是一个元组，元组包含两个元素:
                第一个元素，字符串类型，代表数据名称(path_key形式)
                第二个元素，列表，代表数据
        """
        x_pattern = re.compile(x_select)
        y_pattern = re.compile(y_select)

        x = None
        x_key = None
        ys = []
        for key, value in self.items():
            if x_pattern.match(key):
                assert x is None, "x has been set, duplicated x, " \
                                  "%s [stored] vs %s [store]" % (
                                      x_key, key
                                  )
                x_key = key
                x = value
            else:
                if y_pattern.match(key):
                    ys.append((key, value))

        return (x_key, x), ys


def method_compare(method_args, x_select='iteration',
                   y_select='accuracy|prf_avg_.*|.*loss.*', last_only=True):
    """
    @ Dev

    对比不同方法在不同指标下的表现，两种呈现方式，
    一种以表的形式，一种以图的形式
    在图形式下，如果有多个指标，将会绘制多个子图
    # todo 统一相同label的线的颜色
    # todo 处理x_select出的x长度不一致的case

    Parameters
    ----------
    method_args: list[tuple(method_name, filename)]
    x_select: str
    y_select: str
    last_only: bool
        只返回每个指标的最终值
        是的话，以pandas.DataFrame的形式返回一个表，
        # todo 如果没有 pandas 包, 将返回一个字典，
        此时 x_select 将被无视
        否则，进行可视化

    Returns
    -------

    """
    import json
    key_results = {}

    x_key = None
    x = None
    for model_name, filename in method_args:
        result = ResultAnalyser()
        with open(filename) as f:
            for line in f:
                result.add_record(json.loads(line))
        (x_key, x), ys = result.visual_select(x_select, y_select)
        for y_key, y in ys:
            y_key = y_key.replace('prf_avg_', '')
            if y_key not in key_results:
                key_results[y_key] = {}
            if last_only:
                key_results[y_key][model_name] = y[-1]
            else:
                key_results[y_key][model_name] = y
    if last_only:
        import pandas
        return pandas.DataFrame.from_dict(key_results)

    figure = plt.figure()
    row, col = bd(len(key_results))
    axes = figure.subplots(row, col, sharex='all').flat
    for i, (key, class_values) in enumerate(key_results.items()):
        for class_name, class_value in class_values.items():
            axes[i].plot(x, class_value, label=class_name)
        axes[i].set_title(key)
        axes[i].set_xlabel(x_key)
        handles, labels = axes[i].get_legend_handles_labels()
        lines, labels = zip(*sorted(zip(handles, labels), key=lambda x: x[1]))
        axes[i].legend(lines, labels)
    figure.tight_layout()
    plt.show()


def universe(result, select='accuracy|prf_avg_.*|.*loss.*'):
    """
    平均指标可视化

    Parameters
    ----------
    result: ResultAnalyser

    select: str

    """
    (x_key, x), ys = result.visual_select(y_select=select)
    plt.figure()
    plt.title('universe')
    for y_key, y in ys:
        plt.plot(x, y, label=y_key.split('_')[-1])
        # todo 添加最终结果的线annotation
        # plt.plot([0, x[-1]], [y[-1], y[-1]], '--')
        # todo 添加收敛位置的线annotation

    plt.xlabel(x_key)

    plt.legend()

    plt.show()


def bd(num):
    """
    合数最大分解

    Parameters
    ----------
    num: int

    Returns
    -------

    """
    for i in range(int(math.ceil(math.sqrt(num))), 1, -1):
        if num % i == 0:
            return i, num // i
    return 1, num


def prf(result, class_num, figure_max_class=8, class_key_map=None):
    """
    @分类问题
    分别显示每个类的prf图
    每个子图包含三条线：precision, recall, f1

    Parameters
    ----------
    result: ResultAnalyser
        结果记录
    class_num: int
        类别总数
    figure_max_class: int
        每一幅图可包含的最大子图数
    class_key_map: dict{int: *}
        类名映射字典
    """
    import math

    def key_name(key):
        if class_key_map is not None:
            return class_key_map[key]
        else:
            return str(key)

    (x_key, x), yss = result.visual_selects(
        y_selects=[r'prf_%s_.*' % class_id for class_id in range(class_num)])
    figure_max_class = class_num if not figure_max_class else figure_max_class
    row, col = bd(figure_max_class)
    figures = [plt.figure(i) for i in
               range(int(math.floor(class_num / figure_max_class)))]
    subplots = [figures[i].subplots(row, col, sharex='all', squeeze=False).flat
                for i in range(len(figures))]
    for ys in yss:
        for y_key, y in ys:
            class_id = int(ys[0][0].split('_')[1])
            sp = subplots[class_id // figure_max_class][
                class_id % figure_max_class]
            sp.set_title('class_' + key_name(class_id))
            sp.plot(x, y, label=y_key.split('_')[-1])
            plt.xlabel(x_key)

    for sps in subplots:
        for sp in sps:
            handles, labels = sp.get_legend_handles_labels()
            sp.legend(handles, labels)

    for figure in figures:
        figure.tight_layout()
    plt.show()


def evaluation_group(result, select):
    """

    Parameters
    ----------
    result: ResultAnalyser
    select: str

    Returns
    -------

    """
    # todo 统一不同子图间相同label的颜色
    (x_key, x), ys = result.visual_select(y_select=select)
    plt.figure()
    plt.title('precision')
    for y_key, y in ys:
        label = y_key.split('_')[1]
        plt.plot(x, y, label=label)

    plt.xlabel(x_key)
    plt.ylabel(r'%')

    axes = plt.gca()
    handles, labels = axes.get_legend_handles_labels()
    lines, labels = zip(*sorted(zip(handles, labels), key=lambda x: int(x[1])))
    plt.legend(lines, labels, ncol=math.ceil(len(labels) / 8))
    plt.show()


def precision_group(result):
    """

    Parameters
    ----------
    result: ResultAnalyser

    """
    evaluation_group(result, "prf_\d+_precision")


def recall_group(result):
    """

    Parameters
    ----------
    result: ResultAnalyser

    """
    evaluation_group(result, "prf_\d+_precision")


def f1_group(result):
    """

    Parameters
    ----------
    result: ResultAnalyser

    """
    evaluation_group(result, "prf_\d+_f1")


def pandas_api(result):
    # todo use pandas axes to form the graph
    """try to plot the a1 and a2 on the same figure,
    but now will have three figure, and target figure empty"""
    import pandas
    import numpy as np
    (x_key, x), yss = result.visual_select()
    column, datas = zip(*yss)
    f = plt.figure()
    a1 = pandas.DataFrame.from_records(
        np.asarray(list(datas)).T, columns=column
    )
    b = a1.plot()
    f.axes.append(b)
    a2 = pandas.DataFrame.from_records(
        np.asarray(list(datas)).T, columns=column
    )
    b = a2.plot()
    f.axes.append(b)
    f.canvas.draw()
    plt.show()


if __name__ == '__main__':
    print(method_compare(
        method_args=[('m1', 'result1.json'), ('m2', 'result2.json')]
    ))
    method_compare(
        method_args=[('m1', 'result1.json'), ('m2', 'result2.json')],
        last_only=False
    )



# coding: utf-8
# create by tongshiwei on 2019/4/8

__all__ = ["AttrDict"]


class AttrDict(dict):
    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, item):
        return self[item]

    def __delattr__(self, item):
        del self[item]



# coding: utf-8
# Copyright @tongshiwei

from __future__ import absolute_import
from __future__ import print_function

import datetime
import pathlib

from mxnet import cpu

import longling.ML.MxnetHelper.glue.parser as parser
from longling.ML.MxnetHelper.glue.parser import path_append, var2exp, eval_var
from longling.ML.MxnetHelper.toolkit.optimizer_cfg import get_optimizer_cfg, \
    get_update_steps
from longling.ML.MxnetHelper.toolkit.select_exp import all_params as _select
from longling.lib.utilog import config_logging, LogLevel


class Configuration(parser.Configuration):
    # 目录配置
    model_name = str(pathlib.Path(__file__).parents[1].name)

    root = pathlib.Path(__file__).parents[2]
    dataset = ""
    timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
    workspace = ""

    root_data_dir = "$root/data/$dataset" if dataset else "$root/data"
    data_dir = "$root_data_dir/data"
    root_model_dir = "$root_data_dir/model/$model_name"
    model_dir = "$root_model_dir/$workspace" if workspace else root_model_dir
    cfg_path = "$model_dir/configuration.json"

    root = str(root)
    root_data_dir = str(root_data_dir)
    root_model_dir = str(root_model_dir)

    # 训练参数设置
    begin_epoch = 0
    end_epoch = 100
    batch_size = 32
    save_epoch = 1

    # 优化器设置
    optimizer, optimizer_params = get_optimizer_cfg(name="base")
    lr_params = {
        "step": 100,
        "max_update_steps": get_update_steps(
            update_epoch=10,
            batches_per_epoch=1000,
        ),
    }

    # 更新保存参数，一般需要保持一致
    train_select = _select
    save_select = train_select

    # 运行设备
    ctx = cpu()

    # 用户变量
    # 超参数
    hyper_params = {}

    # 说明
    caption = ""

    def __init__(self, params_json=None, **kwargs):
        """
        Configuration File, including categories:

        * directory setting
        * optimizer setting
        * training parameters
        * equipment
        * parameters saving setting
        * user parameters

        Parameters
        ----------
        params_json: str
            The path to configuration file which is in json format
        kwargs: dict
            Parameters to be reset.
        """
        super(Configuration, self).__init__(
            logger=config_logging(
                logger=self.model_name,
                console_log_level=LogLevel.INFO
            )
        )

        params = self.class_var
        if params_json:
            params.update(self.load_cfg(cfg_path=params_json))
        params.update(**kwargs)

        for param, value in params.items():
            setattr(self, "%s" % param, value)

        # set dataset
        if kwargs.get("dataset"):
            kwargs["root_data_dir"] = "$root/data/$dataset"
        # set workspace
        if kwargs.get("workspace"):
            kwargs["model_dir"] = "$root_model_dir/$workspace"

        # rebuild relevant directory or file path according to the kwargs
        _dirs = [
            "workspace", "root_data_dir", "data_dir", "root_model_dir",
            "model_dir"
        ]
        for _dir in _dirs:
            exp = var2exp(
                kwargs.get(_dir, getattr(self, _dir)),
                env_wrap=lambda x: "self.%s" % x
            )
            setattr(self, _dir, eval(exp))

        _vars = [
            "ctx"
        ]
        for _var in _vars:
            if _var in kwargs:
                try:
                    setattr(self, _var, eval_var(kwargs[_var]))
                except TypeError:
                    pass

        self.validation_result_file = path_append(
            self.model_dir, "result.json", to_str=True
        )
        self.cfg_path = path_append(
            self.model_dir, "configuration.json", to_str=True
        )

    def dump(self, cfg_path=None, override=False):
        cfg_path = self.cfg_path if cfg_path is None else cfg_path
        super(Configuration, self).dump(cfg_path, override)

    @staticmethod
    def load(cfg_path, **kwargs):
        return Configuration(Configuration.load_cfg(cfg_path, **kwargs))


class ConfigurationParser(parser.ConfigurationParser):
    pass


def directory_check(class_obj):
    print("data_dir", class_obj.data_dir)
    print("model_dir", class_obj.model_dir)


if __name__ == '__main__':
    # Advise: firstly checkout whether the directory is correctly (step 1) and
    # then generate the paramters configuation file
    # to check the details (step 2)

    # step 1
    directory_check(Configuration)

    # 命令行参数配置
    _kwargs = ConfigurationParser.get_cli_cfg(Configuration)

    cfg = Configuration(
        **_kwargs
    )
    print(_kwargs)
    print(cfg)

    # # step 2
    cfg.dump(override=True)
    try:
        logger = cfg.logger
        cfg.load_cfg(cfg.cfg_path)
        cfg.logger = logger
        cfg.logger.info('format check done')
    except Exception as e:
        print("parameters format error, may contain illegal data type")
        raise e



# coding: utf-8
# create by tongshiwei on 2018/6/23


class Tree(object):
    def __init__(self, value, *args, **kwargs):
        self.value = value

    @property
    def children(self):
        raise NotImplementedError


class BinaryTree(Tree):
    def __init__(self, value=None, left_child=None, right_child=None):
        super(BinaryTree, self).__init__(value)
        self.left_child = left_child
        self.right_child = right_child

    def set_left_child(self, left_child):
        self.left_child = left_child

    def set_right_child(self, right_child):
        self.right_child = right_child

    @property
    def children(self):
        return [self.left_child, self.right_child]


class MultiChildrenTree(Tree):
    def __init__(self, value=None, children=None):
        super(MultiChildrenTree, self).__init__(value)
        self.__children = children

    @property
    def children(self):
        return self.__children



# coding: utf-8
# create by tongshiwei on 2019/4/7

from mxnet import gluon

__all__ = ["PairwiseLoss"]


class PairwiseLoss(gluon.loss.Loss):
    def __init__(self, weight=None, batch_axis=-1, margin=0, **kwargs):
        super(PairwiseLoss, self).__init__(weight, batch_axis, **kwargs)
        self.margin = margin

    def hybrid_forward(self, F, pos, neg, *args, **kwargs):
        loss = F.add_n(F.negative(pos), neg)
        loss = F.add(loss, self.margin)
        sym = F.relu(loss)
        loss = F.MakeLoss(sym)
        return loss



# coding:utf-8
# created by tongshiwei on 2018/7/31
"""
我忘记这个是写来干嘛用的了……先留着吧
"""
from mxnet import ndarray, symbol

__all__ = ["PVWeight"]


class PVWeight(object):
    def __init__(self, params, var):
        self.params = params
        self.var = var

    def __call__(self, F):
        if F is ndarray:
            return self.params.data()
        elif F is symbol:
            return self.var
        else:
            return None



# coding: utf-8
# create by tongshiwei on 2017/12/2
from __future__ import absolute_import

string_types = str,
integer_types = int


def unistr(x):
    return x


def tostr(x):
    return str(x)

# py2 aborted
# string_types = basestring,
# integer_types = (int, long)
# unistr = lambda x: x.decode('utf-8')
# tostr = lambda x: unicode(x)



# coding: utf-8

"""此模块用以进行流处理"""
# todo 完善注释

from __future__ import print_function

import codecs
import os
import sys

from longling.base import string_types

__all__ = ['rf_open', 'wf_open', 'wf_close', 'flush_print', 'json_load',
           'pickle_load', 'AddPrinter']


class StreamError(Exception):
    pass


def flush_print(*values, **kwargs):
    sep = kwargs.get('sep', "")
    end = kwargs.get('end', "")
    print('\r', *values, sep=sep, end=end, flush=True)


def build_dir(path, mode=0o775):
    """
    创建目录，从path中解析出目录路径，如果目录不存在，创建目录

    Parameters
    ----------
    path: str
    mode: int

    """
    dirname = os.path.dirname(path)
    if not dirname or os.path.exists(dirname):
        return
    os.makedirs(dirname, mode)


def check_file(path, size=None):
    """
    检查文件是否存在，size给定时，检查文件大小是否一致

    Parameters
    ----------
    path: str
    size: int

    Returns
    -------
    file exist or not: bool

    """
    if os.path.exists(path):
        return size == os.path.getsize(path) if size is not None else True
    return False


def rf_open(filename, encoding='utf-8', **kwargs):
    if sys.version_info[0] == 3:
        return open(filename, encoding=encoding, **kwargs)
    else:
        return open(filename, **kwargs)


def json_load(fp, **kwargs):
    import json
    if isinstance(fp, string_types):
        fp = rf_open(fp, **kwargs)
        datas = json.load(fp, **kwargs)
        fp.close()
        return datas
    else:
        return json.load(fp, **kwargs)


def pickle_load(fp, encoding='utf-8', mode='rb', **kwargs):
    import pickle
    if isinstance(fp, string_types):
        fp = open(fp, mode=mode, **kwargs)
        datas = pickle.load(fp, encoding=encoding, **kwargs)
        fp.close()
        return datas
    else:
        return pickle.load(fp, encoding=encoding, **kwargs)


def wf_open(stream_name='', mode="w", encoding="utf-8"):
    r"""
    Simple wrapper to codecs for writing.

    stream_name为空时 mode - w 返回标准错误输出 stderr;
    否则，返回标准输出 stdout

    stream_name不为空时，返回文件流

    Parameters
    ----------
    stream_name: str or None
    mode: str
    encoding: str
        编码方式，默认为 utf-8
    Returns
    -------
    write_stream: StreamReaderWriter
        返回打开的流
    """
    if not stream_name:
        if mode == "w":
            return sys.stderr
        else:
            return sys.stdout
    elif isinstance(stream_name, string_types):
        build_dir(stream_name)
        if mode == "wb":
            return open(stream_name, mode=mode)
        return codecs.open(stream_name, mode=mode, encoding=encoding)
    else:
        raise StreamError("wf_open: %s" % stream_name)


def wf_close(stream):
    if stream in {sys.stdin, sys.stdout, sys.stderr}:
        return True
    else:
        try:
            stream.close()
        except Exception:
            raise StreamError('wf_close: %s' % stream)


class AddPrinter(object):
    def __init__(self, fp, values_wrapper, **kwargs):
        self.fp = fp
        self.value_wrapper = values_wrapper
        self.kwargs = kwargs

    def add(self, value):
        print(self.value_wrapper(value), file=self.fp, **self.kwargs)



from __future__ import print_function

import os

from longling.base import string_types
from longling.lib.stream import wf_open
from longling.lib.utilog import config_logging, LogLevel

__all__ = ["requirements2list"]

logger = config_logging(
    logger="requires_install", console_log_level=LogLevel.INFO
)

REQUIRES_FILE = "requires.txt"


def requirements2list(requirements_file):
    return open(requirements_file).read().split("\n")


def __find_file(root, name):
    files_path = set()
    for relpath, dirs, files in os.walk(root):
        if name in files:
            full_path = os.path.join(relpath, name)
            files_path.add(os.path.normpath(os.path.abspath(full_path)))
            logger.debug("found %s" % full_path)
    return files_path


def __sperator_replace(input_str: string_types):
    if not input_str:
        return os.path.join(".", os.path.sep)
    if '.' in input_str:
        return input_str.replace(".", os.path.sep)
    return input_str


def run(module_names=None, default_confirm=True, user_mode=True, source=None,
        rfile_path="requires.txt"):
    """

    Parameters
    ----------
    module_names
        root module names, can be string or list,
        when not specified, that is None, use "longling"
    default_confirm
        tag to indicating whether or not to
        confirm every requires.txt installed
    user_mode
        whether to use "--user" parameter in pip cmd
    source
        can be http://pypi.mirrors.ustc.edu.cn/simple
    rfile_path
        when specified, output the result to the file. default "requires.txt"

    Returns
    -------

    """
    assert module_names is None or type(module_names) is list or type(
        module_names) in string_types
    if module_names is None:
        module_names = "longling"
    if type(module_names) in string_types:
        module_names = [module_names]
    logger.info("require file path is %s%s" % (
        rfile_path,
        " all depending packages will be installed" if not rfile_path else ""))
    logger.info(
        "Confirm Mode: %s\nthese module requires.txt will be check: %s" % (
            "all confirm without any more permission" if default_confirm
            else "each requires file will need confirmation before installed",
            module_names))
    logger.info("user mode: %s" % user_mode)
    logger.info("specify source: %s" % "Auto" if not source else source)
    files_path = set()
    package_set = set()
    for module_name in module_names:
        module_name = module_name.strip("longling.")
        path = __sperator_replace(module_name)
        logger.info("searching %s" % path)
        new_files = __find_file(path, REQUIRES_FILE)
        for new_file in new_files:
            if new_file not in files_path:
                if default_confirm or input(
                        "install all package in %s?(Y/N)\n" % new_file) in (
                        'Y', 'y', ''):
                    files_path.add(new_file)
    logger.info("%s requires files have been found:\n\t%s" % (
        len(files_path), "\n\t".join(files_path)))
    for file_name in files_path:
        try:
            with open(file_name) as re_file:
                for line in re_file:
                    if line.strip():
                        package_set.add(line.strip())
        except Exception as e:
            logger.error(e)
    logger.info("%s packages will be installed:\n\t%s" % (
        len(package_set), "\n\t".join(package_set)))
    if not rfile_path and (package_set and (
            default_confirm or input("installed? (Y/N)\n") in ('Y', 'y', ''))):
        logger.info("installing")
        install_command = "pip install -r {}"
        if user_mode:
            install_command += " --user"
        if source:
            install_command += " -i %s" % source
        for file_name in files_path:
            os.system(install_command.format(file_name))
            # print(install_command.format(file_name))

    if rfile_path and package_set:
        logger.info("output required packages to %s" % rfile_path)
        with wf_open(rfile_path) as wf:
            for package in package_set:
                print(package, file=wf)


if __name__ == '__main__':
    print(requirements2list("doc/requirements"))



# coding: utf-8
# create by tongshiwei on 2019/4/7

from . import attention, highway, normalization, sequence

from .attention import *
from .highway import *
from .normalization import *
from .sequence import *

__all__ = attention.__all__ + highway.__all__ \
          + normalization.__all__ + sequence.__all__



# coding: utf-8
# create by tongshiwei on 2019/4/7
import math

import mxnet as mx
from mxnet import gluon

__all__ = ["TransE"]


class TransE(gluon.HybridBlock):
    def __init__(self,
                 entities_size, relations_size, dim=50,
                 d='L1', **kwargs):
        super(TransE, self).__init__(**kwargs)
        self.d = d

        with self.name_scope():
            self.entity_embedding = gluon.nn.Embedding(
                entities_size, dim,
                weight_initializer=mx.init.Uniform(6 / math.sqrt(dim))
            )

            self.relation_embedding = gluon.nn.Embedding(
                relations_size, dim,
                weight_initializer=mx.init.Uniform(6 / math.sqrt(dim))
            )

            self.batch_norm = gluon.nn.BatchNorm()

    def hybrid_forward(self, F, sub, rel, obj, **kwargs):
        sub = self.entity_embedding(sub)
        rel = self.relation_embedding(rel)
        obj = self.entity_embedding(obj)

        sub = F.L2Normalization(sub)
        rel = F.L2Normalization(rel)
        obj = F.L2Normalization(obj)

        sub = self.batch_norm(sub)
        rel = self.batch_norm(rel)
        obj = self.batch_norm(obj)
        distance = F.add_n(sub, rel, F.negative(obj))
        if self.d == 'L2':
            return F.norm(distance, axis=1)
        elif self.d == 'L1':
            return F.sum(F.abs(distance), axis=1)



from __future__ import print_function

import pathlib
import argparse
import os

from longling.base import string_types
from longling.lib.utilog import config_logging, LogLevel

logger = config_logging(
    logger="dependency", console_log_level=LogLevel.INFO
)

REQUIREMENTS_FILE = "requirements"


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--module_name", required=False,
        nargs='+',
        default=None,
    )
    args = parser.parse_args()
    find_dependency(args.module_name)


def requirements2list(requirements_file):
    return [
        elem for elem in open(requirements_file).read().split("\n") if elem
    ]


def __find_file(root, name):
    files_path = set()
    for relpath, dirs, files in os.walk(root):
        if name in files:
            full_path = os.path.join(relpath, name)
            files_path.add(os.path.normpath(os.path.abspath(full_path)))
            logger.debug("found %s" % full_path)
    return files_path


def __sperator_replace(input_str: string_types):
    if not input_str:
        return os.path.join(".", os.path.sep)
    if '.' in input_str:
        return input_str.replace(".", os.path.sep)
    return input_str


def find_dependency(module_names=None):
    """

    Parameters
    ----------
    module_names
        root module names, can be string or list,
        when not specified, that is None, use "longling"

    Returns
    -------

    """
    assert module_names is None or type(module_names) is list or type(
        module_names) in string_types
    if module_names is None:
        module_names = "longling"
    if type(module_names) in string_types:
        module_names = [module_names]
    logger.info(
        "These module %s will be check: %s" % (
            REQUIREMENTS_FILE,
            module_names
        )
    )
    root = pathlib.PurePath(__file__).parent
    files_path = set()
    package_set = set()
    for module_name in module_names:
        module_name = module_name.strip("longling.")
        path = root
        if module_name:
            path = root / __sperator_replace(module_name)
        logger.info("searching %s" % path)
        new_files = __find_file(path, REQUIREMENTS_FILE)
        for new_file in new_files:
            if new_file not in files_path:
                files_path.add(new_file)
    logger.info(
        "%s requires files have been found:\n\t%s" % (
            len(files_path), "\n\t".join(files_path))
    )
    for file_name in files_path:
        try:
            for req in requirements2list(file_name):
                package_set.add(req)
        except Exception as e:
            logger.error(e)
    package_set -= {"longling"}
    logger.info(
        "%s dependency packages have be founded:\n\t%s" % (
            len(package_set), ",".join(package_set)
        )
    )

    return package_set


if __name__ == '__main__':
    logger.setLevel(LogLevel.WARN)
    print("\n".join(find_dependency()))



# coding: utf-8
# create by tongshiwei on 2018/6/24
from longling.lib.utilog import config_logging, LogLevel

logger = config_logging(logger="graph", console_log_level=LogLevel.INFO)


class AddList(list):
    def add(self, elem):
        return self.append(elem)



import re


class CommentFilter(object):
    # 评论内容过滤器，过滤 只含表情，数字 以及 回复（含有 //@ ）评论
    def __init__(self):
        self.emoji_re = re.compile('['
                                   '\U0001F300-\U0001F64F'
                                   '\U0001F680-\U0001F6FF'
                                   '\u2600-\u2B55]+',
                                   re.UNICODE)

    def is_valid(self, data):
        if re.sub("^\d+$", "", self.emoji_re.sub("", data['text'])):
            if "//@" not in data['text']:
                return True
        return False



# coding: utf-8
# create by tongshiwei on 2018/7/14

import math

try:
    NAN = math.nan
except (AttributeError, ImportError):
    NAN = float('nan')

__all__ = ["MovingLoss"]


class LossMonitor(object):
    def __init__(self, loss_function_names, *args, **kwargs):
        self.losses = {name: NAN for name in loss_function_names}

    def __str__(self):
        return str(self.losses)

    def update(self, *args, **kwargs):
        raise NotImplementedError

    def reset(self):
        for name in self.losses:
            self.losses[name] = NAN

    def values(self):
        return self.losses.values()

    def items(self):
        return self.losses.items()

    def keys(self):
        raise self.losses.keys()


class MovingLoss(LossMonitor):
    def __init__(self, loss_function_names, smoothing_constant=0.01):
        super(MovingLoss, self).__init__(loss_function_names)
        self.smoothing_constant = smoothing_constant

    def update(self, name, loss_value):
        self.losses[name] = (
            loss_value if math.isnan(self.losses[name])
            else (1 - self.smoothing_constant) * self.losses[
                name] + self.smoothing_constant * loss_value
        )



# coding: utf-8

"""日志设定文件"""

import logging

from longling.base import string_types
from longling.lib.stream import build_dir

__all__ = ["LogLevel", "config_logging"]


class LogLevel(object):
    ERROR = logging.ERROR
    WARN = logging.WARN
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    CRITICAL = logging.CRITICAL


def config_logging(filename=None,
                   log_format='%(name)s, %(levelname)s %(message)s',
                   level=logging.INFO,
                   logger=None, console_log_level=None, propagate=False,
                   mode='a',
                   file_format=None):
    """
    主日志设定文件

    Parameters
    ----------
    filename: str or None
        日志存储文件名，不为空时将创建文件存储日志
    log_format: str
        默认日志输出格式
    level: int
        默认日志等级
    logger: str or logging.logger
        日志logger名，可以为空（使用root logger），
        字符串类型（创建对应名logger），logger
    console_log_level: int or None
        屏幕日志等级，不为空时，使能屏幕日志输出
    propagate: bool
    mode: str
    file_format: str or None
        文件日志输出格式，为空时，使用log_format
    Returns
    -------

    """
    if logger is None:
        logger = logging.getLogger()
    elif isinstance(logger, string_types):
        logger = logging.getLogger(logger)
    # need a clean state, for some module may
    # have called logging functions already (i.e. logging.info)
    # in that case, a default handler would been appended,
    # causing undesired output to stderr
    for handler in logger.handlers:
        logger.removeHandler(handler)
    formatter = logging.Formatter(log_format)
    logger.setLevel(level)
    if not propagate:
        logger.propagate = False
    if filename and filename is not None:
        build_dir(filename)
        handler = logging.FileHandler(filename, mode=mode)
        file_formatter = formatter
        if file_format:
            file_formatter = logging.Formatter(file_format)
        handler.setFormatter(file_formatter)
        logger.addHandler(handler)
    if console_log_level is not None:
        ch = logging.StreamHandler()
        ch.setFormatter(logging.Formatter(log_format))
        ch.setLevel(console_log_level)
        logger.addHandler(ch)

    return logger



# coding: utf-8
# create by tongshiwei on 2019/4/7

from mxnet import gluon

__all__ = ["HighwayCell"]


class HighwayCell(gluon.HybridBlock):
    def __init__(self, dim, highway_layer_activation='relu',
                 highway_trans_layer_activation=None, prefix=None,
                 params=None):
        super(HighwayCell, self).__init__(prefix, params)
        highway_trans_layer_activation = highway_trans_layer_activation \
            if highway_trans_layer_activation else highway_layer_activation
        with self.name_scope():
            self.high_fc = gluon.nn.Dense(
                dim,
                activation=highway_layer_activation
            )
            self.high_trans_fc = gluon.nn.Dense(
                dim,
                activation=highway_trans_layer_activation
            )

    def hybrid_forward(self, F, x, *args, **kwargs):
        high_fc = self.high_fc(x)
        high_trans_fc = self.high_trans_fc(x)
        return high_fc * high_trans_fc + x * (1 - high_trans_fc)



# coding: utf-8
# create by tongshiwei on 2019/4/7

import mxnet as mx
from mxnet import symbol, ndarray

__all__ = ["getF", "copy_net"]


def getF(input):
    r"""
    判断输入参数类型

    Parameters
    ----------
    input

    Returns
    -------

    """
    if isinstance(input, symbol.Symbol):
        return symbol
    elif isinstance(input, ndarray.NDArray):
        return ndarray
    else:
        raise TypeError(
            "the type of input should be either Symbol or NDArray, now is %s"
            % type(input)
        )


def copy_net(src_net, target_net, select=None):
    """
    复制网络

    Parameters
    ----------
    src_net
    target_net
    select

    Returns
    -------

    """
    src_params = src_net.collect_params(select=select)
    target_params = target_net.collect_params(select=select)

    for name, value in src_params._params.items():
        value = value.data()
        target_params._params[
            name.replace(src_params.prefix, target_params.prefix)
        ].set_data(value)


def get_fine_tune_model(symbol, label, arg_params, num_classes,
                        layer_name='flatten0'):
    """
    Only for static newwotk

    Do not use

    Parameters
    ----------
    symbol:
        the pretrained network symbol
    label:

    arg_params:
        the argument parameters of the pretrained model
    num_classes: int
        the number of classes for the fine-tune datasets
    layer_name: str
        the layer name before the last fully-connected layer
    Returns
    -------

    """
    all_layers = symbol.get_internals()
    net = all_layers[layer_name + '_output']
    net = mx.symbol.FullyConnected(
        data=net, num_hidden=num_classes, name='fc1'
    )
    net = mx.symbol.SoftmaxOutput(data=net, label=label, name='softmax')
    new_args = dict({k: arg_params[k] for k in arg_params if 'fc1' not in k})
    return net, new_args



# coding: utf-8
# created by tongshiwei on 17-12-17

from __future__ import absolute_import

from longling.base import *

from longling.lib.stream import *

from longling.lib import clock
from longling.lib import process_pattern
from longling.lib import stream
from longling.lib import utilog
from longling.lib import candylib

__version__ = '1.0.0'



# coding: utf-8
# create by tongshiwei on 2019/4/12
from copy import deepcopy

import mxnet as mx

__all__ = ["get_optimizer_cfg", "get_lr_scheduler", "get_update_steps"]

optimizers = {
    "base": (
        "Adam",
        {
            "learning_rate": 10e-3,
            "wd": 10e-4,
            "clip_gradient": 1,
        }
    ),
    "recurrent": (
        "RMSProp",
        {
            "learning_rate": 10e-4,
            "wd": 10e-4,
            "clip_gradient": 1,
        }
    ),
    "stable": (
        "sgd",
        {
            "learning_rate": 10e-2,
            "wd": 10e-4,
            "clip_gradient": 1,
            "momentum": 0.9,
        }
    ),
    "word_embedding": (
        "adagrad",
        {
            "learning_rate": 10e-2,
            "wd": 10e-4,
            "clip_gradient": 1,
        }
    )
}


def get_update_steps(update_epoch, batches_per_epoch):
    return update_epoch * batches_per_epoch


def get_lr_scheduler(learning_rate, step=None, max_update_steps=None,
                     discount=0.01):
    factor = pow(discount, step / max_update_steps)
    return mx.lr_scheduler.FactorScheduler(
        step, factor,
        stop_factor_lr=learning_rate * discount
    )


def get_optimizer_cfg(name):
    try:
        optimizer, optimizer_params = optimizers[name]
    except KeyError:
        raise KeyError("the name should be in: %s" % ", ".join(optimizers))
    optimizer_params = deepcopy(optimizer_params)

    return optimizer, optimizer_params


def plot_schedule(schedule_fn, iterations=1500):
    import matplotlib.pyplot as plt
    # Iteration count starting at 1
    iterations = [i + 1 for i in range(iterations)]
    lrs = [schedule_fn(i) for i in iterations]
    plt.scatter(iterations, lrs)
    plt.xlabel("Iteration")
    plt.ylabel("Learning Rate")
    plt.show()



# coding: utf-8
# create by tongshiwei on 2019/6/27

from tests.lib.stream import NullDevice
from longling.ML.toolkit.dataset import file_dataset_split, DatasetSplitter


def test_dataset_split():
    pesudo_data = list(range(0, 1000))
    spliter = DatasetSplitter()
    spliter(
        pesudo_data,
        train_buffer=NullDevice(), valid_buffer=NullDevice(), test_buffer=NullDevice(),
        silent=False,
    )



# coding: utf-8
# create by tongshiwei on 2019/4/16

import pathlib

from longling.requires_install import requirements2list

__all__ = ["__dependency__"]

__dependency__ = requirements2list(
    pathlib.PurePath(__file__).with_name("requirements")
)



# coding: utf-8
# created by tongshiwei on 18-2-5

import codecs
import json
import logging
import re

from longling.base import string_types
from longling.lib.stream import wf_open

__all__ = ["EvalFormatter", "MultiClassEvalFormatter"]


def _to_dict(name_value):
    return dict(name_value) if isinstance(name_value, tuple) else name_value


class EvalFormatter(object):
    def __init__(self, logger=logging.getLogger(), dump_file=None, **kwargs):
        self.logger = logger
        if dump_file is not None and isinstance(dump_file, string_types):
            # clean file
            wf_open(dump_file, **kwargs).close()
        self.log_f = dump_file

    @staticmethod
    def _loss_format(name, value):
        return "%s: %s" % (name, value)

    def loss_format(self, loss_name_value):
        msg = "Loss - "
        for name, value in loss_name_value.items():
            msg += self._loss_format(name, value)
        return msg, loss_name_value

    @staticmethod
    def _eval_format(name, value):
        return "Evaluation %s: %s" % (name, value)

    def eval_format(self, eval_name_value):
        msg = []
        for name, value in eval_name_value.items():
            msg.append(self._eval_format(name, value))
        msg = "\t".join([m for m in msg if m])
        data = eval_name_value
        return msg, data

    def __call__(self, tips=None,
                 epoch=None, train_time=None, loss_name_value=None,
                 eval_name_value=None,
                 extra_info=None,
                 dump=True,
                 *args, **kwargs
                 ):

        msg = []
        data = {}

        if tips is not None:
            msg.append("%s" % tips)

        if epoch is not None:
            msg.append("Epoch [%d]:" % epoch)
            data['Epoch'] = epoch

        if train_time is not None:
            msg.append("Train Time-%.3fs" % train_time)
            data['train_time'] = train_time

        if loss_name_value is not None:
            loss_name_value = _to_dict(loss_name_value)
            assert isinstance(
                loss_name_value, dict
            ), "loss_name_value should be None, dict or tuple, " \
               "now is %s" % type(loss_name_value)
            _msg, _data = self.loss_format(loss_name_value)

            msg.append(
                _msg
            )
            data.update(
                _data
            )

        if extra_info is not None:
            extra_info = _to_dict(extra_info)
            assert isinstance(
                extra_info, dict
            ), "extra_info should be None, dict or tuple, " \
               "now is %s" % type(extra_info)
            msg.append(extra_info.items())
            data.update(extra_info)

        msg = ["\t".join([m for m in msg if m])]

        if eval_name_value is not None:
            eval_name_value = _to_dict(eval_name_value)
            assert isinstance(
                eval_name_value, dict
            ), "eval_name_value should be None, dict or tuple, " \
               "now is %s" % type(eval_name_value)
            _msg, _data = self.eval_format(eval_name_value)
            msg.append(
                _msg
            )
            data.update(
                _data
            )

        msg = "\n".join([m for m in msg if m])

        if dump:
            logger = kwargs.get('logger', self.logger)
            logger.info(msg)
            log_f = kwargs.get('log_f', self.log_f)
            if log_f is not None:
                try:
                    if log_f is not None and isinstance(log_f, string_types):
                        log_f = codecs.open(log_f, "a", encoding="utf-8")
                        print(json.dumps(data, ensure_ascii=False), file=log_f)
                        log_f.close()
                    else:
                        print(json.dumps(data, ensure_ascii=False), file=log_f)
                except Exception as e:
                    logger.warning(e)
        return msg, data


class MultiClassEvalFormatter(EvalFormatter):
    def eval_format(self, eval_name_value):
        msg = []
        data = {}

        multi_class_pattern = re.compile(".+_\d+")

        prf = {}
        eval_ids = set()

        for name, value in sorted(eval_name_value.items()):
            if multi_class_pattern.match(name) is not None:
                try:
                    eval_id, class_id = name.split("_")
                except ValueError:
                    continue
                if class_id not in prf:
                    prf[class_id] = {}
                prf[class_id][eval_id] = value
                eval_ids.add(eval_id)
            else:
                msg.append(self._eval_format(name, value))
                data[name] = value

        msg = "\t".join([m for m in msg if m])
        if msg:
            msg += '\n'
        if prf:
            avg = {eval_id: [] for eval_id in eval_ids}
            for class_id in [str(k) for k in
                             sorted([int(k) for k in prf.keys()])]:
                for eval_id, values in avg.items():
                    values.append(prf[class_id][eval_id])
                msg += "--- Category %s" % class_id
                msg_res = sorted(prf[class_id].items(), reverse=True)
                msg += ("\t{}={:.10f}" * len(prf[class_id])).format(
                    *sum(msg_res, ())) + "\n"
            avg = {
                eval_id: sum(values) / len(values)
                for eval_id, values in avg.items()
            }
            msg += "--- Category_Avg "
            msg_res = sorted(avg.items(), reverse=True)
            msg += ("\t{}={:.10f}" * len(avg)).format(*sum(msg_res, ()))
            prf['avg'] = avg
            data['prf'] = prf

        return msg, data


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.INFO)
    formatter = MultiClassEvalFormatter()
    print(formatter(
        eval_name_value={
            "Acuuracy": 0.5,
            "precision_1": 10, "precision_0": 20,
            "recall_0": 1, "recall_1": 2
        }
    )[0])



# coding: utf-8
# create by tongshiwei on 2019/4/13

from mxnet.gluon.nn import Embedding
from mxnet.ndarray import NDArray

__all__ = ["set_embedding_weight"]


def set_embedding_weight(embedding_layer, weight):
    """
    Parameters
    ----------
    embedding_layer: Embedding
    weight: NDArray
    """
    embedding_layer.weight.set_data(weight)



# coding: utf-8
# create by tongshiwei on 2019/4/7

import mxnet as mx
from mxnet import gluon

__all__ = ["get_l2_embedding_weight"]


def get_l2_embedding_weight(F, embedding_size, batch_size=None):
    entries = list(range(embedding_size))
    embedding_weight = []
    batch_size = batch_size if batch_size else embedding_size
    for entity in gluon.data.DataLoader(gluon.data.ArrayDataset(entries),
                                        batch_size=batch_size,
                                        shuffle=False):
        embedding_weight.extend(
            mx.nd.L2Normalization(F(entity)).asnumpy().tolist())

    return embedding_weight



# coding: utf-8
from __future__ import absolute_import

import logging
import time
from contextlib import contextmanager

from longling.lib.utilog import config_logging

_logger = config_logging(logger="clock", console_log_level=logging.INFO)

__all__ = ["Clock", "print_time"]


@contextmanager
def print_time(task, logger=_logger):
    """

    Parameters
    ----------
    task: str
    logger: logging.Logger

    Examples
    --------
    >>> with print_time("task_name"):
    ...     a = 1 + 1

    """
    start_time = time.time()
    logger.info('Starting to %s', task)
    yield
    logger.info('Finished to {} in {:.6f} seconds'.format(
        task,
        time.time() - start_time)
    )


class Clock(object):
    r"""
    计时器

    Parameters
    ----------
    store_dict: dict or None
    logger: logging.logger
    tips: str

    Examples
    --------
    >>> with Clock():
    ...     a = 1 + 1
    """

    def __init__(self, store_dict=None, logger=_logger, tips=''):
        assert store_dict is None or type(store_dict) is dict
        self.process_st = 0
        self.process_et = 0
        self.wall_st = 0
        self.wall_et = 0
        self.store_dict = store_dict
        self.logger = logger
        self.tips = tips

    def start(self):
        """
        开始计时
        Returns
        -------

        """
        self.process_st = time.clock()
        self.wall_st = time.time()
        return self.process_st

    def end(self, wall=False):
        """
        计时结束，返回间隔时间

        Parameters
        ----------
        wall

        Returns
        -------

        """
        self.process_et = time.clock()
        self.wall_et = time.time()
        if wall:
            return self.wall_time
        else:
            return self.process_time

    @property
    def wall_time(self):
        """
        获取程序运行时间（包括等待时间）

        Returns
        -------

        """
        return self.wall_et - self.wall_st

    @property
    def process_time(self):
        """
        获取程序运行时间（不包括等待时间）
        Returns
        -------

        """
        return self.process_et - self.process_st

    def __enter__(self):
        if self.tips:
            self.logger.info(self.tips)
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.logger.info(
            '%ss' % self.end() if not self.tips else '%s %ss' % (
                self.tips, self.end()
            )
        )
        if self.store_dict is not None:
            self.store_dict['wall_time'] = self.wall_time
            self.store_dict['process_time'] = self.process_time


if __name__ == '__main__':
    with Clock():
        time.sleep(1)



# coding:utf-8
# created by tongshiwei on 2018/10/9

import os
import random

from tqdm import tqdm

from longling.lib.stream import wf_open, AddPrinter

__all__ = ["DatasetSplitter", "file_dataset_split"]


class DatasetSplitter(object):
    @staticmethod
    def split(iterable,
              valid_ratio=0.0, test_ratio=0.2,
              train_buffer=None, valid_buffer=None, test_buffer=None, silent=True):
        counter = {
            "train": 0,
            "valid": 0,
            "test": 0,
        }
        for element in tqdm(iterable, disable=silent):
            random_seed = random.random()
            if random_seed < test_ratio:
                buffer = test_buffer
            elif random_seed < valid_ratio + test_ratio:
                buffer = valid_buffer
            else:
                buffer = train_buffer
            buffer.add(element)

        return counter

    def __call__(self, iterable,
                 valid_ratio=0.0, test_ratio=0.2,
                 train_buffer=None, valid_buffer=None, test_buffer=None, silent=True):
        DatasetSplitter.split(
            iterable, valid_ratio, test_ratio,
            train_buffer, valid_buffer, test_buffer,
            silent
        )


def file_dataset_split(filename, valid_ratio=0.0, test_ratio=0.5, root_dir=None,
                       train_filename=None, valid_filename=None, test_filename=None,
                       silent=True):
    if train_filename or test_filename or valid_filename:
        assert train_filename and valid_filename and test_filename
    else:
        train_filename = filename + ".train"
        valid_filename = filename + ".valid"
        test_filename = filename + ".test"

    if root_dir:
        train_filename = os.path.join(root_dir, train_filename)
        test_filename = os.path.join(root_dir, test_filename)

    splitter = DatasetSplitter()

    with open(filename) as f, \
            wf_open(train_filename) as train_wf, \
            wf_open(valid_filename) as valid_wf, \
            wf_open(test_filename) as test_wf:

        train_buffer = AddPrinter(train_wf)
        valid_buffer = AddPrinter(valid_wf)
        test_buffer = AddPrinter(test_wf)

        counter = splitter(
            f, valid_ratio, test_ratio,
            train_buffer, valid_buffer, test_buffer,
            silent,
        )

    return counter



# coding:utf-8
# created by tongshiwei on 2018/10/9

import os
import random

from tqdm import tqdm

from longling.lib.stream import wf_open

__all__ = ["train_test_split"]


def train_test_split(filename, test_ratio=0.5, root_dir=None,
                     train_filename=None, test_filename=None,
                     progress_viz=True):
    if train_filename or test_filename:
        assert train_filename and test_filename
    else:
        train_filename = filename + ".train"
        test_filename = filename + ".test"

    if root_dir:
        train_filename = os.path.join(root_dir, train_filename)
        test_filename = os.path.join(root_dir, test_filename)

    train_cnt = 0
    test_cnt = 0
    with open(filename) as f, wf_open(train_filename) as train_wf, wf_open(
            test_filename) as test_wf:
        for line in tqdm(f, disable=not progress_viz):
            if random.random() < test_ratio:
                wf = test_wf
                test_cnt += 1
            else:
                wf = train_wf
                train_cnt += 1
            print(line, end='', file=wf)

    return train_cnt, test_cnt



# coding:utf-8
# created by tongshiwei on 2018/7/11
"""
This file contains some loss function
"""
import mxnet as mx
from mxnet import gluon

__all__ = ["TimeSeriesPickLoss"]


class TimeSeriesPickLoss(gluon.HybridBlock):
    """
    Loss from Deep Knowledge Tracing

    Notes
    -----
    The loss has been average, so when call the step method of trainer,
    batch_size should be 1
    """

    def hybrid_forward(self, F, pred, pick_index, label, label_mask, *args,
                       **kwargs):
        pred = F.slice(pred, (None, None), (None, -1))
        pred = F.pick(pred, pick_index)
        loss = F.LogisticRegressionOutput(pred, label)
        loss = F.sum(mx.nd.squeeze(
            F.SequenceMask(F.expand_dims(loss, 0), sequence_length=label_mask,
                           use_sequence_length=True)), axis=-1)
        loss = F.mean(loss)
        return loss



# coding: utf-8
# create by tongshiwei on 2019/4/12
import os
import re

import mxnet as mx
from mxnet import gluon, nd

__all__ = ["Module"]


class Module(object):
    def __init__(self, parameters):
        self.cfg = parameters

    def __str__(self):
        """
        显示模块参数
        Display the necessary params of this Module

        Returns
        -------

        """
        string = ["Params"]
        for k, v in vars(self.cfg).items():
            string.append("%s: %s" % (k, v))
        return "\n".join(string)

    @staticmethod
    def load_net(filename, net, ctx=mx.cpu(), allow_missing=False,
                 ignore_extra=False):
        """
        Load the existing net parameters

        Parameters
        ----------
        filename: str
            The model file
        net: HybridBlock
            The network which has been initialized or
            loaded from the existed model
        ctx: Context or list of Context
                Defaults to ``mx.cpu()``.
        allow_missing: bool
        ignore_extra: bool

        Returns
        -------
        The initialized net
        """
        # 根据文件名装载已有的网络参数
        if not os.path.isfile(filename):
            raise FileExistsError
        net.load_parameters(filename, ctx, allow_missing=allow_missing,
                            ignore_extra=ignore_extra)
        return net

    def load(self, net, epoch, ctx=mx.cpu(), allow_missing=False,
             ignore_extra=False):
        """"
        Load the existing net parameters

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized
            or loaded from the existed model
        epoch: str or int
            The epoch which specify the model
        ctx: Context or list of Context
                Defaults to ``mx.cpu()``.
        allow_missing: bool
        ignore_extra: bool

        Returns
        -------
        The initialized net
        """
        # 根据起始轮次装载已有的网络参数
        filename = self.epoch_params_filename(epoch)
        return self.load_net(
            filename, net, ctx, allow_missing=allow_missing,
            ignore_extra=ignore_extra
        )

    def epoch_params_filename(self, epoch):
        raise NotImplementedError

    @staticmethod
    def net_initialize(
            net, model_ctx, initializer=mx.init.Normal(sigma=.1), select=None
    ):
        """初始化网络参数"""
        net.collect_params(select).initialize(initializer, ctx=model_ctx)

    @staticmethod
    def get_trainer(net, optimizer='sgd', optimizer_params=None, select=None):
        """把优化器安装到网络上"""
        trainer = gluon.Trainer(
            net.collect_params(select), optimizer, optimizer_params
        )
        return trainer

    @staticmethod
    def save_params(filename, net, select):
        params = net._collect_params_with_prefix()
        if select:
            pattern = re.compile(select)
            params = {name: value for name, value in params.items() if
                      pattern.match(name)}
        arg_dict = {key: val._reduce() for key, val in params.items()}
        nd.save(filename, arg_dict)



# coding: utf-8
# create by tongshiwei on 2017/11/17

from __future__ import absolute_import

from . import candylib
from . import clock
from . import stream
from . import utilog
from . import process_pattern



# coding:utf-8
# created by tongshiwei on 2018/11/8
import numpy as np


class Factor(object):
    def __init__(
            self, initial_value, end_value, change_rate=0, warmup_steps=0
    ):
        self.cnt = 0
        self._initial_value = initial_value
        self._end_value = end_value
        self._sign = int(np.sign(initial_value - end_value))
        self._value = initial_value
        self.warmup_steps = warmup_steps
        self._change_rate = - self._sign * change_rate

    @property
    def value(self):
        if self._sign == 0:
            return self._value
        else:
            if self.cnt >= self.warmup_steps and self._sign * (
                    self._value - self._end_value) > 0:
                self._value += self._change_rate
            else:
                self.cnt += 1

    def reset(self):
        self._value = self._initial_value
        self.cnt = 0


def linearly_decay(initial_rate, min_rate, period, warmup=0):
    assert initial_rate > min_rate
    return initial_rate, min_rate, (initial_rate - min_rate) / (
            period - warmup
    )



# coding: utf-8
# create by tongshiwei on 2019/4/12

from mxnet import gluon
from tqdm import tqdm

__all__ = ["extract", "transform", "etl"]


def extract(data_src):
    raise NotImplementedError


def transform(raw_data, params):
    # 定义数据转换接口
    # raw_data --> batch_data

    batch_size = params.batch_size

    return gluon.data.DataLoader(gluon.data.ArrayDataset(raw_data), batch_size)


def etl(*args, **kwargs):
    raw_data = extract(*args)
    transform(raw_data, **kwargs)
    raise NotImplementedError


if __name__ == '__main__':
    for data in tqdm(extract("")):
        pass



# coding:utf-8
# created by tongshiwei on 2018/8/2

from mxnet import symbol, ndarray
from mxnet.gluon.parameter import tensor_types

from longling.lib.candylib import as_list

__all__ = ["format_sequence", "mask_sequence_variable_length"]


def format_sequence(length, inputs, layout, merge, in_layout=None):
    """

    Parameters
    ----------
    length
    inputs
    layout
    merge
    in_layout

    Returns
    -------

    """
    assert inputs is not None, \
        "unroll(inputs=None) has been deprecated. " \
        "Please create input variables outside unroll."

    axis = layout.find('T')
    batch_axis = layout.find('N')
    batch_size = 0
    in_axis = in_layout.find('T') if in_layout is not None else axis
    if isinstance(inputs, symbol.Symbol):
        F = symbol
        if merge is False:
            assert len(inputs.list_outputs()) == 1, \
                "unroll doesn't allow grouped symbol as input. " \
                "Please convert " \
                "to list with list(inputs) first or " \
                "let unroll handle splitting."
            inputs = list(symbol.split(
                inputs, axis=in_axis, num_outputs=length, squeeze_axis=1
            ))
    elif isinstance(inputs, ndarray.NDArray):
        F = ndarray
        batch_size = inputs.shape[batch_axis]
        if merge is False:
            assert length is None or length == inputs.shape[in_axis]
            inputs = as_list(ndarray.split(inputs, axis=in_axis,
                                           num_outputs=inputs.shape[in_axis],
                                           squeeze_axis=1))
    else:
        assert length is None or len(inputs) == length
        if isinstance(inputs[0], symbol.Symbol):
            F = symbol
        else:
            F = ndarray
            batch_size = inputs[0].shape[batch_axis]
        if merge is True:
            inputs = F.stack(*inputs, axis=axis)
            in_axis = axis

    if isinstance(inputs, tensor_types) and axis != in_axis:
        inputs = F.swapaxes(inputs, dim1=axis, dim2=in_axis)

    return inputs, axis, F, batch_size


def mask_sequence_variable_length(F, data, length, valid_length, time_axis,
                                  merge):
    """

    Parameters
    ----------
    F
    data
    length
    valid_length
    time_axis
    merge

    Returns
    -------

    """
    assert valid_length is not None
    if not isinstance(data, tensor_types):
        data = F.stack(*data, axis=time_axis)
    outputs = F.SequenceMask(
        data, sequence_length=valid_length,
        use_sequence_length=True,
        axis=time_axis
    )
    if not merge:
        outputs = as_list(
            F.split(
                outputs, num_outputs=length, axis=time_axis, squeeze_axis=True
            )
        )
    return outputs



from __future__ import absolute_import

import re

from .process_pattern_base import *

mode_dict = {}


def register(number):
    def _register(func):
        if number in mode_dict:
            logger.warning(
                "mode-%s %s existed, overriding by %s"
                % (number, mode_dict[number].__name__, func.__name__)
            )
        else:
            logger.info("register mode-%s %s" % (number, func.__name__))
        mode_dict[number] = func
        return func

    return _register


@register(0)
def _init_patterns0(line, pps=None):
    try:
        line = line.decode('utf-8').strip()
    except AttributeError:
        pass
    try:
        line1, line2 = line.split(':')
    except Exception:
        raise ProcessPatternLineError('error in %s' % line)

    v = line1.strip()
    ps1 = []
    if v:
        try:
            p = re.compile(v)
        except Exception:
            raise ProcessPatternLineError('error in %s' % line)
        ps1.append(p)

    v = line2.strip()
    assert v
    ps2 = []
    try:
        p = re.compile(v)
    except Exception:
        raise ProcessPatternLineError('error in %s' % line)
    ps2.append(p)

    if pps is not None and isinstance(pps, dict):
        pps[line] = (ps1, ps2)
    return line, (ps1, ps2)


@register(1)
def _init_patterns1(line, pps):
    try:
        line = line.decode('utf-8').strip()
    except AttributeError:
        pass
    try:
        line1, line2 = line.split(':')
    except Exception:
        raise ProcessPatternLineError('error in %s' % line)

    vs = line1.split()
    ps1 = []
    for v in vs:
        try:
            p = re.compile(v)
        except Exception:
            raise ProcessPatternLineError('error in %s' % line)
        ps1.append(p)

    vs = line2.split()
    ps2 = []
    for v in vs:
        try:
            p = re.compile(v)
        except Exception:
            print("error:", v)
            raise ProcessPatternLineError('error in %s' % line)
        ps2.append(p)

    if pps is not None and isinstance(pps, dict):
        pps[line] = (ps1, ps2)

    return line, (ps1, ps2)


def _init_patterns(location, patterns):
    """
    创建一个模式字典

    Parameters
    ----------
    location: str
        正则规则文件位置
    patterns:

    Returns
    -------
    dict
        模式字典
    """
    pps = {}
    with open(location) as f:
        for line in f:
            try:
                patterns(line, pps)
            except ProcessPatternLineError as e:
                print(e)
                continue
    return pps


def init_patterns(location, mode):
    return _init_patterns(location, mode_dict[mode])


def line_init_patterns(line, mode, pps=None):
    """

    Parameters
    ----------
    line
    mode: int
    pps

    Returns
    -------

    """
    if mode not in mode_dict:
        raise ProcessPatternNotExistedPatternError(
            "available mode is %s" % list(mode_dict.keys())
        )
    return mode_dict[mode](line, pps)



# coding: utf-8
# create by tongshiwei on 2019/4/7

from mxnet import gluon

__all__ = ["TextCNN"]


class TextCNN(gluon.HybridBlock):
    def __init__(self, sentence_size, vec_size, channel_size=None,
                 num_output=2, filter_list=(1, 2, 3, 4), num_filters=60,
                 dropout=0.0, batch_norm=True, activation="tanh",
                 pool_type='max',
                 **kwargs):
        """
        TextCNN 模型，for 2D and 3D

        Parameters
        ----------
        sentence_size: int
            句子长度
        vec_size: int
            特征向量维度
        channel_size: int or None
            if None, use Text2D
            int 则为 channel（特征向量二维维度）
        num_output: int
            输出维度
        filter_list: Iterable
            The output dimension for each convolutional layer according
            to the filter sizes,
            which are the number of the filters learned by the layers.
        num_filters: int or Iterable
            The size of each convolutional layer, int or Iterable.
            When Iterable, len(filter_list) equals to the number of
            convolutional layers.
        dropout: float
        batch_norm: bool
        activation: str
        pool_type: str
        kwargs
        """
        super(TextCNN, self).__init__(**kwargs)
        self.sentence_size = sentence_size
        self.vec_size = vec_size
        self.channel_size = channel_size
        self.num_output = num_output
        self.filter_list = filter_list
        if isinstance(num_filters, int):
            self.num_filters = [num_filters] * len(self.filter_list)
        assert len(self.filter_list) == len(self.num_filters)
        self.batch_norm = batch_norm
        self.activation = activation

        self.conv = [0] * len(self.filter_list)
        self.pool = [0] * len(self.filter_list)
        self.bn = [0] * len(self.filter_list)

        pool2d = gluon.nn.MaxPool2D if pool_type == "max" \
            else gluon.nn.AvgPool2D

        pool3d = gluon.nn.MaxPool3D if pool_type == "max" \
            else gluon.nn.AvgPool3D

        with self.name_scope():
            for i, (filter_size, num_filter) in enumerate(
                    zip(self.filter_list, self.num_filters)):
                conv = gluon.nn.Conv2D(
                    num_filter,
                    kernel_size=(filter_size, self.vec_size),
                    activation=self.activation) if not self.channel_size else \
                    gluon.nn.Conv3D(
                        num_filter,
                        kernel_size=(
                            filter_size, self.vec_size, self.channel_size
                        ),
                        activation=activation
                    )
                setattr(self, "conv%s" % i, conv)

                pool = pool2d(
                    pool_size=(self.sentence_size - filter_size + 1, 1),
                    strides=(1, 1)) if not self.channel_size else pool3d(
                    pool_size=(self.sentence_size - filter_size + 1, 1, 1),
                    strides=(1, 1, 1))
                setattr(self, "pool%s" % i, pool)

                if self.batch_norm:
                    setattr(self, "bn%s" % i, gluon.nn.BatchNorm())

            self.dropout = gluon.nn.Dropout(dropout)

            self.fc = gluon.nn.Dense(num_output)

    def hybrid_forward(self, F, x, *args, **kwargs):
        x = F.expand_dims(x, axis=1)

        pooled_outputs = []
        for i, _ in enumerate(self.filter_list):
            convi = getattr(self, "conv%s" % i)(x)
            pooli = getattr(self, "pool%s" % i)(convi)
            if self.batch_norm:
                pooli = getattr(self, "bn%s" % i)(pooli)
            pooled_outputs.append(pooli)

        total_filters = sum(self.num_filters)
        concat = F.Concat(dim=1, *pooled_outputs)
        h_pool = F.Reshape(data=concat, shape=(0, total_filters))
        h_drop = self.dropout(h_pool)

        fc = self.fc(h_drop)

        return fc



# coding: utf-8
# create by tongshiwei on 2019/4/11

import argparse
import inspect
import json
import logging
import os
import re
from pathlib import PurePath

from longling import wf_open

__all__ = [
    "CLASS_EXCLUDE_NAMES", "get_class_var",
    "get_parsable_var", "load_configuration_json",
    "var2exp", "path_append",
    "Configuration", "ConfigurationParser"
]

CLASS_EXCLUDE_NAMES = set(vars(object).keys()) | {
    '__module__', '__main__', '__dict__', '__weakref__',
}


def get_class_var(class_type, exclude_names=None):
    default_excluded_names = CLASS_EXCLUDE_NAMES if exclude_names is None \
        else exclude_names
    excluded_names = default_excluded_names if not hasattr(
        class_type, "excluded_names"
    ) else class_type.excluded_names() | default_excluded_names

    variables = {
        k: v for k, v in vars(class_type).items() if
        not inspect.isroutine(v) and k not in excluded_names
    }
    return variables


def parse_params(params, parse_functions=None):
    if parse_functions is not None:
        for name, func in parse_functions.items():
            if name in params:
                params[name] = func(params[name])
    return params


def get_parsable_var(class_obj, parse_exclude=None, dump_parse_functions=None):
    params = get_class_var(class_obj, exclude_names=parse_exclude)
    return parse_params(params, dump_parse_functions)


def load_configuration_json(fp, load_parse_function=None):
    params = json.load(fp)
    return parse_params(params, load_parse_function)


def var2exp(var_str, env_wrap=lambda x: x):
    """
    将含有 $ 标识的变量转换为表达式

    Parameters
    ----------
    var_str
    env_wrap

    Examples
    --------
    >>> root = "dir"
    >>> dataset = "d1"
    >>> eval(var2exp("$root/data/$dataset"))
    'dir/data/d1'
    """
    var_str = str(var_str)

    pattern = re.compile(r"\$(\w+)")

    env_vars = pattern.findall(var_str)

    exp = """str("%s").format(%s)""" % (
        re.sub(r"\$\w+", "{}", var_str), ",".join(
            [env_wrap(env_var) for env_var in env_vars]
        )
    )
    return exp


def path_append(path, addition=None, to_str=False):
    """

    Parameters
    ----------
    path: str or PurePath
    addition: str or PurePath
    to_str: bool
        Convert the new path to str
    Returns
    -------

    """
    path = PurePath(path)
    new_path = path / addition if addition else path
    if to_str:
        return str(new_path)
    return new_path


class Configuration(object):
    def __init__(self, logger=logging):
        self.logger = logger

    @property
    def class_var(self):
        """
        获取所有设定的参数

        Returns
        -------
        parameters: dict
            all variables used as parameters
        """
        return get_class_var(self)

    @property
    def parsable_var(self):
        """
        获取可以进行命令行设定的参数

        Returns
        -------
        store_vars: dict
            可以进行命令行设定的参数
        """
        return get_parsable_var(
            self,
            parse_exclude=None,
            dump_parse_functions=None,
        )

    @staticmethod
    def load_cfg(params_json, **kwargs):
        with open(params_json) as f:
            params = load_configuration_json(
                f, load_parse_function=None
            )
        params.update(kwargs)
        return params

    @staticmethod
    def load(cfg_path, **kwargs):
        Configuration(Configuration.load_cfg(cfg_path, **kwargs))

    def dump(self, cfg_json, override=False):
        if os.path.isfile(cfg_json) and not override:
            self.logger.warning("file %s existed, dump aborted" % cfg_json)
            return
        self.logger.info("writing configuration parameters to %s" % cfg_json)
        with wf_open(cfg_json) as wf:
            json.dump(self.parsable_var, wf, indent=4)

    def __str__(self):
        return str(self.parsable_var)

    def __getitem__(self, item):
        return getattr(self, item)

    def items(self):
        return {k: v for k, v in self.parsable_var}

    @staticmethod
    def excluded_names():
        """
        获取非参变量集

        Returns
        -------
        exclude names set: set
            所有非参变量
        """
        return CLASS_EXCLUDE_NAMES | {
            'class_var', 'parsable_var', 'items', 'load', 'dump'
        }


def value_parse(value):
    r"""
    将含有关键字的字符串转换为关键字指定的类型
    支持的关键字类型为python的基础数据类型:
    int, float, dict, list, set, tuple, None

    Parameters
    ----------
    value: str
        字符串值
    Examples
    --------
    >>> value_parse("int(1.0)")
    1
    >>> value_parse("float(1.0)")
    1.0
    >>> value_parse("dict(a=1, b=2.0, c='d')")
    {'a': 1, 'b': 2.0, 'c': 'd'}
    >>> value_parse("list([1, 2, 3, 4])")
    [1, 2, 3, 4]
    >>> value_parse("set([1, 1, 2, 3])")
    {1, 2, 3}
    >>> value_parse("tuple([1, 2, 3])")
    (1, 2, 3)
    >>> value_parse("None") is None
    True
    """
    if re.findall(r"(int|float|dict|list|set|tuple)\(.*\)|None", value):
        value = eval(value)
    return value


def parse_dict_string(string):
    if not string:
        return None
    else:
        name_value_items = [
            item.strip().split("=") for item in string.strip().split(",")
        ]
        return {
            name_value[0]: value_parse(name_value[1])
            for name_value in name_value_items
        }


def args_zips(args=None, defaults=None):
    args = [] if args is None else args
    defaults = [] if defaults is None else defaults
    assert len(args) >= len(defaults)
    zipped_args = []
    zipped_args.extend([
        {"name": arg} for arg in args[:len(args) - len(defaults)]
    ])
    zipped_args.extend([
        {"name": arg, "default": default} for arg, default in zip(
            args[len(args) - len(defaults):], defaults
        )
    ])
    return zipped_args


class ConfigurationParser(argparse.ArgumentParser):
    def __init__(self, class_obj, excluded_names=None, *args, **kwargs):
        excluded_names = {
            'logger'
        } if excluded_names is None else excluded_names

        super(ConfigurationParser, self).__init__(*args, **kwargs)
        params = {k: v for k, v in get_class_var(class_obj).items()}
        for param, value in params.items():
            if param in excluded_names:
                continue
            if isinstance(value, dict):
                format_tips = ", dict variables, " \
                              "use format: <key>=<value>(,<key>=<value>)"
                self.add_argument(
                    '--%s' % param,
                    help='set %s, default is %s%s' % (
                        param, value, format_tips
                    ), default=value,
                    type=parse_dict_string,
                )
            else:
                self.add_argument(
                    '--%s' % param,
                    help='set %s, default is %s' % (param, value),
                    default=value,
                    type=value_parse,
                )
        self.add_argument(
            '--kwargs', required=False,
            help=r"add extra argument here, "
                 r"use format: <key>=<value>(,<key>=<value>)"
        )
        self.sub_command_parsers = None

    @staticmethod
    def parse(arguments):
        arguments = vars(arguments)
        args_dict = dict()
        for k, v in arguments.items():
            if k == "kwargs" and v:
                kwargs = parse_dict_string(v)
                if kwargs:
                    args_dict.update(kwargs)
                continue
            args_dict[k] = v
        return args_dict

    @staticmethod
    def get_cli_cfg(params_class):
        params_parser = ConfigurationParser(params_class)
        kwargs = params_parser.parse_args()
        kwargs = params_parser.parse(kwargs)
        return kwargs

    def __call__(self, args=None):
        kwargs = self.parse_args(args)
        kwargs = self.parse(kwargs)
        return kwargs

    def add_subcommand(self, command_parameters):
        if self.sub_command_parsers is None:
            self.sub_command_parsers = self.add_subparsers(
                parser_class=argparse.ArgumentParser,
                help='help for sub-command'
            )
        subparsers = self.sub_command_parsers
        subcommand, parameters = command_parameters
        subparser = subparsers.add_parser(
            subcommand, help="%s help" % subcommand
        )
        subparser.set_defaults(subcommand=subcommand)
        for parameter in parameters:
            if "default" in parameter:
                subparser.add_argument(
                    "--%s" % parameter["name"],
                    default=parameter["default"],
                    required=False,
                    type=value_parse,
                    help="set %s, default is %s" % (
                        parameter["name"], parameter["default"]
                    )
                )
            else:
                subparser.add_argument(
                    parameter["name"],
                    type=value_parse,
                    help="set %s" % parameter["name"]
                )

    @staticmethod
    def func_spec(f):
        argspec = inspect.getfullargspec(f)
        return f.__name__, args_zips(
            argspec.args, argspec.defaults
        ) + args_zips(
            argspec.kwonlyargs,
            argspec.kwonlydefaults.values() if argspec.kwonlydefaults else []
        )


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# coding: utf-8
# Copyright @tongshiwei
"""
This file define the networks structure and
provide a simplest training and testing example.
"""

import logging
import os

import mxnet as mx
from mxnet import gluon, nd, autograd
from tqdm import tqdm

from longling.ML.MxnetHelper.toolkit.ctx import split_and_load
from longling.ML.MxnetHelper.toolkit.viz import plot_network, VizError

# set parameters
try:
    # for python module
    from .etl import transform
    from .configuration import Configuration
except (ImportError, SystemError):
    # for python script
    from etl import transform
    from configuration import Configuration

__all__ = ["NetName", "net_viz", "fit_f", "BP_LOSS_F", "eval_f"]


# todo: define your network symbol here
class NetName(gluon.HybridBlock):
    def __init__(self, **kwargs):
        super(NetName, self).__init__(**kwargs)

        with self.name_scope():
            pass

    def hybrid_forward(self, F, x, *args, **kwargs):
        pass


# todo: optional, visualize the network
def net_viz(_net, _cfg, view_tag=False, **kwargs):
    """visualization check, only support pure static network"""
    batch_size = _cfg.batch_size
    model_dir = _cfg.model_dir
    logger = kwargs.get(
        'logger',
        _cfg.logger if hasattr(_cfg, 'logger') else logging
    )

    try:
        viz_dir = os.path.join(model_dir, "plot/network")
        logger.info("visualization: file in %s" % viz_dir)
        from copy import deepcopy

        viz_net = deepcopy(_net)
        viz_shape = {'data': (batch_size,) + (1,)}
        x = mx.sym.var("data")
        sym = viz_net(x)
        plot_network(
            nn_symbol=sym,
            save_path=viz_dir,
            shape=viz_shape,
            node_attrs={"fixedsize": "false"},
            view=view_tag
        )
    except VizError as e:
        logger.error("error happen in visualization, aborted")
        logger.error(e)


# todo:
def get_data_iter(_cfg):
    def pseudo_data_generation():
        # 在这里定义测试用伪数据流
        import random
        random.seed(10)

        raw_data = [
            [random.random() for _ in range(5)]
            for _ in range(1000)
        ]

        return raw_data

    return transform(pseudo_data_generation(), _cfg)


def fit_f(_net, _data, bp_loss_f, loss_function, loss_monitor):
    data, label = _data
    # todo modify the input to net
    output = _net(data)
    bp_loss = None
    for name, func in loss_function.items():
        # todo modify the input to func
        loss = func(output, label)
        if name in bp_loss_f:
            bp_loss = loss
        loss_value = nd.mean(loss).asscalar()
        if loss_monitor:
            loss_monitor.update(name, loss_value)
    return bp_loss


def eval_f(_net, test_data, ctx=mx.cpu()):
    ground_truth = []
    prediction = []

    def evaluation_function(y_true, y_pred):
        return 0

    for batch_data in tqdm(test_data, "evaluating"):
        ctx_data = split_and_load(
            ctx, *batch_data,
            even_split=False
        )
        for (data, label) in ctx_data:
            output = _net(data)
            pred = output
            ground_truth.extend(label.asnumpy().tolist())
            prediction.extend(pred.asnumpy().tolist())

    return {
        "evaluation_name": evaluation_function(ground_truth, prediction)
    }


BP_LOSS_F = {"L2Loss": gluon.loss.L2Loss}


def numerical_check(_net, _cfg):
    net.initialize()

    datas = get_data_iter(_cfg)

    bp_loss_f = BP_LOSS_F
    loss_function = {}
    loss_function.update(bp_loss_f)
    from longling.ML.toolkit.monitor import MovingLoss
    from longling.ML.MxnetHelper.glue import module

    loss_monitor = MovingLoss(loss_function)

    # train check
    trainer = module.Module.get_trainer(
        _net, optimizer=_cfg.optimizer,
        optimizer_params=_cfg.optimizer_params,
        select=_cfg.train_select
    )

    for epoch in range(0, 100):
        for _data in tqdm(datas):
            with autograd.record():
                bp_loss = fit_f(
                    _net, _data, bp_loss_f, loss_function, loss_monitor
                )
            assert bp_loss is not None
            bp_loss.backward()
            trainer.step(_cfg.batch_size)
        print("epoch-%d: %s" % (epoch, list(loss_monitor.items())))


if __name__ == '__main__':
    cfg = Configuration()

    # generate sym
    net = NetName()

    # # visualiztion check
    # net_viz(net, cfg, False)

    # # numerical check
    # numerical_check(net, cfg)



# coding: utf-8
# create by tongshiwei on 2018/8/12
# Modified from gluonnlp

"""Attention cells."""
from __future__ import absolute_import
from __future__ import print_function

import math

import mxnet as mx
from mxnet.gluon import nn
from mxnet.gluon.block import HybridBlock

__all__ = [
    'AttentionCell',
    'MultiHeadAttentionCell',
    'MLPAttentionCell',
    'DotProductAttentionCell'
]


def _masked_softmax(F, att_score, mask):
    """Ignore the masked elements when calculating the softmax

    Parameters
    ----------
    F : symbol or ndarray
    att_score : Symborl or NDArray
        Shape (batch_size, query_length, memory_length)
    mask : Symbol or NDArray or None
        Shape (batch_size, query_length, memory_length)
    Returns
    -------
    att_weights : Symborl or NDArray
        Shape (batch_size, query_length, memory_length)
    """
    if mask is not None:
        # Fill in the masked scores with a very small value
        mask = F.SequenceMask(F.ones_like(att_score), sequence_length=mask,
                              use_sequence_length=True, axis=1)
        att_score = F.where(mask, att_score, -1e18 * F.ones_like(att_score))
        att_weights = F.softmax(att_score, axis=-1) * mask
    else:
        att_weights = F.softmax(att_score, axis=-1)
    return att_weights


class AttentionCell(HybridBlock):
    """Abstract class for attention cells. Extend the class
     to implement your own attention method.
     One typical usage is to define your own `_compute_weight()`
     function to calculate the weights::

        cell = AttentionCell()
        out = cell(query, key, value, mask)

    """

    def _compute_weight(self, F, query, key, mask=None):
        """Compute attention weights based on the query and the keys

        Parameters
        ----------
        F : symbol or ndarray
        query : Symbol or NDArray
            The query vectors. Shape (batch_size, query_length, query_dim)
        key : Symbol or NDArray
            Key of the memory. Shape (batch_size, memory_length, key_dim)
        mask : Symbol or NDArray or None
            Mask the memory slots.
            Shape (batch_size, query_length, memory_length)
            Only contains 0 or 1
            where 0 means that the memory slot will not be used.
            If set to None. No mask will be used.

        Returns
        -------
        att_weights : Symbol or NDArray
            For single-head attention,
            Shape (batch_size, query_length, memory_length)
            For multi-head attentino,
            Shape (batch_size, num_heads, query_length, memory_length)
        """
        raise NotImplementedError

    def _read_by_weight(self, F, att_weights, value):
        """Read from the value matrix given the attention weights.

        Parameters
        ----------
        F : symbol or ndarray
        att_weights : Symbol or NDArray
            Attention weights.
            For single-head attention,
                Shape (batch_size, query_length, memory_length).
            For multi-head attention,
                Shape (batch_size, num_heads, query_length, memory_length).
        value : Symbol or NDArray
            Value of the memory.
            Shape (batch_size, memory_length, total_value_dim)

        Returns
        -------
        context_vec: Symbol or NDArray
            Shape (batch_size, query_length, context_vec_dim)
        """
        return F.batch_dot(att_weights, value)

    def __call__(self, query, key, value=None,
                 mask=None):  # pylint: disable=arguments-differ
        """Compute the attention.

        Parameters
        ----------
        query : Symbol or NDArray
            Query vector. Shape (batch_size, query_length, query_dim)
        key : Symbol or NDArray
            Key of the memory. Shape (batch_size, memory_length, key_dim)
        value : Symbol or NDArray or None, default None
            Value of the memory. If set to None, the value will be
            set as the key. Shape (batch_size, memory_length, value_dim)
        mask : Symbol or NDArray or None, default None
            Mask of the memory slots.
            Shape (batch_size, query_length, memory_length)
            Only contains 0 or 1
            where 0 means that the memory slot will not be used.
            If set to None. No mask will be used.

        Returns
        -------
        context_vec : Symbol or NDArray
            Shape (batch_size, query_length, context_vec_dim)
        att_weights : Symbol or NDArray
            Attention weights. Shape (batch_size, query_length, memory_length)
        """
        return super(AttentionCell, self).__call__(query, key, value, mask)

    def forward(self, query, key, value=None,
                mask=None):  # pylint: disable=arguments-differ
        if value is None:
            value = key
        if mask is None:
            return super(AttentionCell, self).forward(query, key, value)
        else:
            return super(AttentionCell, self).forward(query, key, value, mask)

    def hybrid_forward(self, F, query, key, value,
                       mask=None):  # pylint: disable=arguments-differ
        att_weights = self._compute_weight(F, query, key, mask)
        context_vec = self._read_by_weight(F, att_weights, value)
        return context_vec, att_weights


class MultiHeadAttentionCell(AttentionCell):
    r"""Multi-head Attention Cell.

    In the MultiHeadAttentionCell, the input query/key/value
     will be linearly projected
    for `num_heads` times with different projection matrices.
    Each projected key, value, query
    will be used to calculate the attention weights and values.
    The output of each head will be
    concatenated to form the final output.

    The idea is first proposed in "[Arxiv2014] Neural Turing Machines" and
    is later adopted in "[NIPS2017] Attention is All You Need" to solve the
    Neural Machine Translation problem.

    Parameters
    ----------
    base_cell : AttentionCell
    query_units : int
        Total number of projected units for query.
        Must be divided exactly by num_heads.
    key_units : int
        Total number of projected units for key.
        Must be divided exactly by num_heads.
    value_units : int
        Total number of projected units for value.
        Must be divided exactly by num_heads.
    num_heads : int
        Number of parallel attention heads
    use_bias : bool, default True
        Whether to use bias when projecting the query/key/values
    weight_initializer : str or `Initializer` or None, default None
        Initializer of the weights.
    bias_initializer : str or `Initializer`, default 'zeros'
        Initializer of the bias.
    prefix : str or None, default None
        See document of `Block`.
    params : str or None, default None
        See document of `Block`.
    """

    def __init__(self, base_cell, query_units, key_units, value_units,
                 num_heads, use_bias=True,
                 weight_initializer=None, bias_initializer='zeros',
                 prefix=None, params=None):
        super(MultiHeadAttentionCell, self).__init__(prefix=prefix,
                                                     params=params)
        self._base_cell = base_cell
        self._query_units = query_units
        self._key_units = key_units
        self._value_units = value_units
        self._num_heads = num_heads
        self._use_bias = use_bias
        if self._query_units % self._num_heads != 0:
            raise ValueError(
                'In MultiHeadAttetion, the query_units '
                'should be divided exactly'
                ' by the number of heads. '
                'Received query_units={}, num_heads={}'.format(
                    key_units, num_heads)
            )

        if self._key_units % self._num_heads != 0:
            raise ValueError(
                'In MultiHeadAttetion, the key_units should be divided exactly'
                ' by the number of heads. '
                'Received key_units={}, num_heads={}'.format(
                    key_units, num_heads
                ))

        if self._value_units % self._num_heads != 0:
            raise ValueError(
                'In MultiHeadAttetion, the value_units '
                'should be divided exactly'
                ' by the number of heads. Received value_units={}, '
                'num_heads={}'.format(value_units, num_heads)
            )

        with self.name_scope():
            self.proj_query = nn.Dense(units=self._query_units,
                                       use_bias=self._use_bias,
                                       flatten=False,
                                       weight_initializer=weight_initializer,
                                       bias_initializer=bias_initializer,
                                       prefix='query_')
            self.proj_key = nn.Dense(units=self._key_units,
                                     use_bias=self._use_bias,
                                     flatten=False,
                                     weight_initializer=weight_initializer,
                                     bias_initializer=bias_initializer,
                                     prefix='key_')
            self.proj_value = nn.Dense(units=self._value_units,
                                       use_bias=self._use_bias,
                                       flatten=False,
                                       weight_initializer=weight_initializer,
                                       bias_initializer=bias_initializer,
                                       prefix='value_')

    def __call__(self, query, key, value=None, mask=None):
        """Compute the attention.

        Parameters
        ----------
        query : Symbol or NDArray
            Query vector. Shape (batch_size, query_length, query_dim)
        key : Symbol or NDArray
            Key of the memory. Shape (batch_size, memory_length, key_dim)
        value : Symbol or NDArray or None, default None
            Value of the memory. If set to None,
            the value will be set as the key.
            Shape (batch_size, memory_length, value_dim)
        mask : Symbol or NDArray or None, default None
            Mask of the memory slots.
            Shape (batch_size, query_length, memory_length)
            Only contains 0 or 1
            where 0 means that the memory slot will not be used.
            If set to None. No mask will be used.

        Returns
        -------
        context_vec : Symbol or NDArray
            Shape (batch_size, query_length, context_vec_dim)
        att_weights : Symbol or NDArray
            Attention weights of multiple heads.
            Shape (batch_size, num_heads, query_length, memory_length)
        """
        return super(MultiHeadAttentionCell, self).__call__(query, key, value,
                                                            mask)

    def _compute_weight(self, F, query, key, mask=None):
        query = self.proj_query(
            query)  # Shape (batch_size, query_length, query_units)
        # Shape (batch_size * num_heads, query_length, ele_units)
        query = F.transpose(query.reshape(shape=(0, 0, self._num_heads, -1)),
                            axes=(0, 2, 1, 3)) \
            .reshape(shape=(-1, 0, 0), reverse=True)
        key = self.proj_key(key)
        key = F.transpose(key.reshape(shape=(0, 0, self._num_heads, -1)),
                          axes=(0, 2, 1, 3)).reshape(shape=(-1, 0, 0),
                                                     reverse=True)
        if mask is not None:
            mask = F.broadcast_axis(F.expand_dims(mask, axis=1),
                                    axis=1, size=self._num_heads) \
                .reshape(shape=(-1, 0, 0), reverse=True)
        att_weights = self._base_cell._compute_weight(F, query, key, mask)
        return att_weights.reshape(shape=(-1, self._num_heads, 0, 0),
                                   reverse=True)

    def _read_by_weight(self, F, att_weights, value):
        att_weights = att_weights.reshape(shape=(-1, 0, 0), reverse=True)
        value = self.proj_value(value)
        value = F.transpose(value.reshape(shape=(0, 0, self._num_heads, -1)),
                            axes=(0, 2, 1, 3)).reshape(shape=(-1, 0, 0),
                                                       reverse=True)
        context_vec = self._base_cell._read_by_weight(F, att_weights, value)
        context_vec = F.transpose(
            context_vec.reshape(shape=(-1, self._num_heads, 0, 0),
                                reverse=True),
            axes=(0, 2, 1, 3)).reshape(shape=(0, 0, -1))
        return context_vec


class MLPAttentionCell(AttentionCell):
    r"""Concat the query and the key and use a single-hidden-layer MLP to
    get the attention score.
    We provide two mode, the standard mode and the normalized mode.

    In the standard mode::

        score = v tanh(W [h_q, h_k] + b)

    In the normalized mode (Same as TensorFlow)::

        score = g v / ||v||_2 tanh(W [h_q, h_k] + b)

    This type of attention is first proposed in

    .. Bahdanau et al., Neural Machine Translation by
    Jointly Learning to Align and Translate.
       ICLR 2015

    Parameters
    ----------
    units : int
    act : Activation, default nn.Activation('tanh')
    normalized : bool, default False
        Whether to normalize the weight that maps the embedded
        hidden states to the final score. This strategy
        can be interpreted as a type of
        "[NIPS2016] Weight Normalization".
    dropout : float, default 0.0
        Attention dropout.
    weight_initializer : str or `Initializer` or None, default None
        Initializer of the weights.
    bias_initializer : str or `Initializer`, default 'zeros'
        Initializer of the bias.
    prefix : str or None, default None
        See document of `Block`.
    params : ParameterDict or None, default None
        See document of `Block`.
    """

    def __init__(self, units, act=nn.Activation('tanh'), normalized=False,
                 dropout=0.0,
                 weight_initializer=None, bias_initializer='zeros',
                 prefix=None, params=None):
        # Define a temporary class to implement the normalized version
        # TODO(sxjscience) Find a better solution
        class _NormalizedScoreProj(HybridBlock):
            def __init__(self, in_units, weight_initializer=None, prefix=None,
                         params=None):
                super(_NormalizedScoreProj, self).__init__(prefix=prefix,
                                                           params=params)
                self.g = self.params.get('g', shape=(1,),
                                         init=mx.init.Constant(
                                             1.0 / math.sqrt(in_units)),
                                         allow_deferred_init=True)
                self.v = self.params.get('v', shape=(1, in_units),
                                         init=weight_initializer,
                                         allow_deferred_init=True)

            def hybrid_forward(self, F, x, g,
                               v):  # pylint: disable=arguments-differ
                v = F.broadcast_div(v, F.sqrt(F.dot(v, v, transpose_b=True)))
                weight = F.broadcast_mul(g, v)
                out = F.FullyConnected(x, weight, None, no_bias=True,
                                       num_hidden=1,
                                       flatten=False, name='fwd')
                return out

        super(MLPAttentionCell, self).__init__(prefix=prefix, params=params)
        self._units = units
        self._act = act
        self._normalized = normalized
        self._dropout = dropout
        with self.name_scope():
            self._dropout_layer = nn.Dropout(dropout)
            self._query_mid_layer = nn.Dense(
                units=self._units, flatten=False,
                use_bias=True,
                weight_initializer=weight_initializer,
                bias_initializer=bias_initializer,
                prefix='query_'
            )
            self._key_mid_layer = nn.Dense(
                units=self._units, flatten=False,
                use_bias=False,
                weight_initializer=weight_initializer,
                prefix='key_'
            )
            if self._normalized:
                self._attention_score = _NormalizedScoreProj(
                    in_units=units,
                    weight_initializer=weight_initializer,
                    prefix='score_'
                )
            else:
                self._attention_score = nn.Dense(
                    units=1, in_units=self._units,
                    flatten=False, use_bias=False,
                    weight_initializer=weight_initializer,
                    prefix='score_'
                )

    def _compute_weight(self, F, query, key, mask=None):
        mapped_query = self._query_mid_layer(query)
        mapped_key = self._key_mid_layer(key)
        mid_feat = F.broadcast_add(F.expand_dims(mapped_query, axis=2),
                                   F.expand_dims(mapped_key, axis=1))
        mid_feat = self._act(mid_feat)
        att_score = self._attention_score(mid_feat).reshape(shape=(0, 0, 0))
        att_weights = self._dropout_layer(_masked_softmax(F, att_score, mask))
        return att_weights


class DotProductAttentionCell(AttentionCell):
    r"""Dot product attention between the query and the key.

    Depending on parameters, defined as::

        units is None:
            score = <h_q, h_k>
        units is not None and luong_style is False:
            score = <W_q h_q, W_k h_k>
        units is not None and luong_style is True:
            score = <W h_q, h_k>

    Parameters
    ----------
    units: int or None, default None
        Project the query and key to vectors with `units` dimension
        before applying the attention. If set to None,
        the query vector and the key vector are directly
        used to compute the attention and
        should have the same dimension::

            If the units is None,
                score = <h_q, h_k>
            Else if the units is not None and luong_style is False:
                score = <W_q h_q, W_k, h_k>
            Else if the units is not None and luong_style is True:
                score = <W h_q, h_k>

    luong_style: bool, default False
        If turned on, the score will be::

            score = <W h_q, h_k>

        `units` must be the same as the dimension of the key vector
    scaled: bool, default True
        Whether to divide the attention weights
        by the sqrt of the query dimension.
        This is first proposed in "[NIPS2017] Attention is all you need."::

            score = <h_q, h_k> / sqrt(dim_q)

    normalized: bool, default False
        If turned on, the cosine distance is used, i.e::

            score = <h_q / ||h_q||, h_k / ||h_k||>

    use_bias : bool, default True
        Whether to use bias in the projection layers.
    dropout : float, default 0.0
        Attention dropout
    weight_initializer : str or `Initializer` or None, default None
        Initializer of the weights
    bias_initializer : str or `Initializer`, default 'zeros'
        Initializer of the bias
    prefix : str or None, default None
        See document of `Block`.
    params : str or None, default None
        See document of `Block`.
    """

    def __init__(self, units=None, luong_style=False, scaled=True,
                 normalized=False, use_bias=True,
                 dropout=0.0, weight_initializer=None,
                 bias_initializer='zeros', prefix=None, params=None):
        super(DotProductAttentionCell, self).__init__(prefix=prefix,
                                                      params=params)
        self._units = units
        self._scaled = scaled
        self._normalized = normalized
        self._use_bias = use_bias
        self._luong_style = luong_style
        self._dropout = dropout
        if self._luong_style:
            assert units is not None, 'Luong style attention ' \
                                      'is not available without explicitly ' \
                                      'setting the units'
        with self.name_scope():
            self._dropout_layer = nn.Dropout(dropout)
        if units is not None:
            with self.name_scope():
                self._proj_query = nn.Dense(
                    units=self._units,
                    use_bias=self._use_bias,
                    flatten=False,
                    weight_initializer=weight_initializer,
                    bias_initializer=bias_initializer,
                    prefix='query_'
                )
                if not self._luong_style:
                    self._proj_key = nn.Dense(
                        units=self._units,
                        use_bias=self._use_bias,
                        flatten=False,
                        weight_initializer=weight_initializer,
                        bias_initializer=bias_initializer,
                        prefix='key_'
                    )

    def _compute_weight(self, F, query, key, mask=None):
        if self._units is not None:
            query = self._proj_query(query)
            if not self._luong_style:
                key = self._proj_key(key)
            elif F == mx.nd:
                assert query.shape[-1] == key.shape[
                    -1], 'Luong style attention requires key to ' \
                         'have the same dim as the projected ' \
                         'query. Received key {}, query {}.'.format(
                    key.shape, query.shape)
        if self._normalized:
            query = F.L2Normalization(query, mode='spatial')
            key = F.L2Normalization(key, mode='spatial')
        if self._scaled:
            query = F.contrib.div_sqrt_dim(query)
        att_score = F.batch_dot(query, key, transpose_b=True)
        att_weights = self._dropout_layer(_masked_softmax(F, att_score, mask))
        return att_weights


class SelfAttentionCell(HybridBlock):
    r"""Self Attention, query, key and value are the same


    This type of attention is first proposed in

    .. Lin Z, Feng M, Santos C N, et al.
    A structured self-attentive sentence embedding[J].
    arXiv preprint arXiv:1703.03130, 2017.

    Parameters
    ----------
    s1_units : int
    s2_units : int
    normalized : bool, default False
        Whether to normalize the weight that maps the embedded
        hidden states to the final score.
        This strategy can be interpreted as a type of
        "[NIPS2016] Weight Normalization".
    dropout : float, default 0.0
        Attention dropout.
    weight_initializer : str or `Initializer` or None, default None
        Initializer of the weights.
    prefix : str or None, default None
        See document of `Block`.
    params : ParameterDict or None, default None
        See document of `Block`.
    """

    def __init__(self, s1_units, s2_units,
                 normalized=False, dropout=0.0,
                 weight_initializer=None, prefix=None, params=None):
        super(SelfAttentionCell, self).__init__(prefix=prefix, params=params)
        self._normalized = normalized
        self._dropout = dropout
        with self.name_scope():
            self._dropout_layer = nn.Dropout(dropout)
            self._s1_layer = nn.Dense(s1_units, flatten=False, use_bias=False,
                                      weight_initializer=weight_initializer)
            self._s2_layer = nn.Dense(s2_units, flatten=False, use_bias=False,
                                      weight_initializer=weight_initializer)
            self._s1_act = nn.Activation('tanh')

    def hybrid_forward(self, F, x, mask=None, **kwargs):
        att_score_s1 = self._s1_act(self._s1_layer(x))
        att_score_s2 = self._s2_layer(att_score_s1)
        att_weights = self._dropout_layer(
            _masked_softmax(F, att_score_s2, mask))
        return att_weights


if __name__ == '__main__':
    from mxnet import nd

    a = nd.ones((128, 10, 7))
    b = nd.ones((128, 10, 7))
    self_attention = SelfAttentionCell(5, 3)
    self_attention.initialize()
    s = self_attention(a)
    print(s.shape)



# coding: utf-8
# create by tongshiwei on 2018/6/23

from __future__ import absolute_import

from .base import *
from .dict import *
from .linkedlist import *
from .linkedlist import DoublyLinkedList as DList, SingleLinkedList as SList
from .tree import *
from .tree import BinaryTree as BTree, MultiChildrenTree as MTree



__all__ = ['as_list']


def as_list(obj):
    r"""A utility function that converts the argument to a list
    if it is not already.

    Parameters
    ----------
    obj : object
        argument to be converted to a list

    Returns
    -------
    list_obj: list
        If `obj` is a list or tuple, return it. Otherwise,
        return `[obj]` as a single-element list.

    Examples
    --------
    >>> as_list(1)
    [1]
    >>> as_list([1])
    [1]
    """
    if isinstance(obj, (list, tuple)):
        return obj
    else:
        return [obj]


def get_all_subclass(cls):
    subclass = set()

    def _get_all_subclass(cls, res_set):
        res_set.add(cls)
        for sub_cls in cls.__subclasses__():
            _get_all_subclass(sub_cls, res_set)

    _get_all_subclass(cls, res_set=subclass)
    return subclass


class RegisterNameExistedError(Exception):
    pass


class Register(object):
    def __init__(self):
        self.reg = set()

    def register(self, name):
        if name not in self.reg:
            self.reg.add(name)
            return True
        else:
            raise RegisterNameExistedError("name %s existed" % name)


if __name__ == '__main__':
    import doctest

    doctest.testmod()



from longling.lib.utilog import config_logging, LogLevel

logger = config_logging(
    logger='process_pattern', console_log_level=LogLevel.WARN
)


class ProcessPatternNotExistedPatternError(Exception):
    pass


class ProcessPatternLineError(Exception):
    pass


class ProcessPatternEncodedError(Exception):
    pass



# coding: utf-8
# create by tongshiwei on 2019/4/7
from mxnet.gluon import utils as gutil

from longling.lib.candylib import as_list

__all__ = ["real_ctx", "split_and_load"]


def real_ctx(ctx, data_len):
    ctx = as_list(ctx)
    if data_len < len(ctx):
        ctx = ctx[:1]
    return ctx


def split_and_load(ctx, *args, **kwargs):
    ctx = real_ctx(ctx, len(args[0]))
    return zip(*[gutil.split_and_load(arg, ctx, **kwargs) for arg in args])



# coding: utf-8
# create by tongshiwei on 2019/4/13
"""
Here are some frequently used regex expression for select in collect_params()
"""
# 包含所有参数
all_params = None

# 除外所有embedding层
block_embedding = "^(?!.*embedding)"



# coding: utf-8
# created by tongshiwei on 17-12-17
import io
import os
import re
from distutils.core import setup

from setuptools import find_packages

test_deps = [
    'pytest>=4',
    'pytest-cov>=2.6.0',
    'pytest-pep8>=1',
]
doc_deps = [
    'sphinx',
    'sphinx-rtd-theme',
    'recommonmark'
]
dev_deps = test_deps + doc_deps + [
    'setuptools>=40',
    'wheel'
]


def read(*names, **kwargs):
    with io.open(
            os.path.join(os.path.dirname(__file__), *names),
            encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


VERSION = find_version('longling', '__init__.py')

setup(
    name='longling',
    version=VERSION,
    author='Sherlock, Shiwei Tong',
    author_email='tongsw@mail.ustc.edu.cn',
    include_package_data=True,
    packages=find_packages(
        include=[
            "longling",
            "*.lib", "*.lib.*",
            "*.ML",
            "*.ML.MxnetHelper*",
            "*.ML.toolkit*"
        ],
        exclude=[
            "*.mx_example", "*.gluon_example*", "*.gluon_exp*",
            "*.mxnet_old*",
        ]
    ),
    entry_points={
        "console_scripts": [
            "glue = longling.ML.MxnetHelper.glue.glue:cli",
            "longling_install = longling.dependency:cli"
        ],
    },
    url='https://gitlab.com/tswsxk/longling.git',
    license='LICENSE.txt',
    description='handy wrapper for many libs',
    long_description=open('README.md', encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "pip"
    ],
    extras_require={
        'test': test_deps,
        'doc': doc_deps,
        'dev': dev_deps
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Development Status :: 4 - Beta",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



# coding: utf-8
# Copyright @tongshiwei
from __future__ import absolute_import

import os

import mxnet as mx
from mxnet import autograd

from longling.ML.MxnetHelper.glue import module
from longling.ML.MxnetHelper.toolkit.ctx import split_and_load
from .configuration import Configuration
from .sym import NetName, fit_f, eval_f

__all__ = ["Module"]


class Module(module.Module):
    """
    模块模板
    train 修改流程

    # 1
    修改 __init__ 和 params 方法
    Module(....) 初始化一些通用的参数，比如模型的存储路径等

    # 2
    定义网络部分，命名为
    sym_gen(), 可以有多个
    也可以使用直接装载的方式生成已有网络

    # 3
    使用可视化方法进行检查网络是否搭建得没有问题
    可视化方法命名为 visualization

    # 4
    定义数据装载部分，可以有多个，命名为
    get_{train/test/data}_iter

    # 5
    定义训练相关的参数及交互提示信息

    # 6
    定义 net_initialize 部分
    定义训练部分

    # 7
    关闭输入输出流

    # optional
    定义 eval 评估模型方法
    定义 load 加载模型方法
    """

    def __init__(self, configuration):
        """

        Parameters
        ----------
        configuration: Configuration

        """
        # 初始化一些通用的参数
        self.cfg = configuration
        self.prefix = os.path.join(
            self.cfg.model_dir, self.cfg.model_name
        )
        self.logger = configuration.logger

        self.sym_gen = NetName

    def dump_configuration(self, filename=None):
        filename = filename if filename is not None \
            else os.path.join(self.cfg.cfg_path)
        self.cfg.dump(filename, override=True)
        return filename

    def epoch_params_filename(self, epoch):
        return self.prefix + "-%04d.parmas" % epoch

    # 部分定义训练相关的方法
    @staticmethod
    def get_trainer(
            net, optimizer='sgd', optimizer_params=None, lr_params=None,
            select=Configuration.train_select
    ):
        if lr_params is not None:
            from longling.ML.MxnetHelper.toolkit.optimizer_cfg import \
                get_lr_scheduler
            optimizer_params["lr_scheduler"] = get_lr_scheduler(**lr_params)
        return module.Module.get_trainer(
            net, optimizer, optimizer_params, select
        )

    def save_params(self, filename, net):
        module.Module.save_params(filename, net, select=self.cfg.save_select)

    def fit(
            self,
            net, begin_epoch, end_epoch, batch_size,
            train_data,
            trainer, bp_loss_f,
            loss_function,
            eval_data=None,
            ctx=mx.cpu(),
            toolbox=None,
            **kwargs
    ):
        """
        API for train

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized or loaded from
            the existed model
        begin_epoch: int
            The begin epoch of this train procession
        end_epoch: int
            The end epoch of this train procession
        batch_size: int
                The size of each batch
        train_data: Iterable
            The data used for this train procession,
            NOTICE: should have been divided to batches
        trainer:
            The trainer used to update the parameters of the net
        bp_loss_f: dict with only one value and one key
            The function to compute the loss for the procession
            of back propagation
        loss_function: dict of function
            Some other measurement in addition to bp_loss_f
        eval_data: Iterable
            The data used for the evaluation at the end of each epoch,
            NOTICE: should have been divided to batches
            Default to ``None``
        ctx: Context or list of Context
            Defaults to ``mx.cpu()``.
        toolbox: dict or None
            Default to ``None``
        kwargs

        Returns
        -------

        """
        # 此方法可以直接使用
        return self.epoch_loop(
            net=net, begin_epoch=begin_epoch, end_epoch=end_epoch,
            batch_size=batch_size,
            train_data=train_data,
            trainer=trainer, bp_loss_f=bp_loss_f,
            loss_function=loss_function,
            test_data=eval_data,
            ctx=ctx,
            toolbox=toolbox,
            **kwargs
        )

    def epoch_loop(
            self,
            net, begin_epoch, end_epoch, batch_size,
            train_data,
            trainer,
            bp_loss_f, loss_function,
            test_data=None,
            ctx=mx.cpu(),
            toolbox=None,
            **kwargs
    ):
        """
        此函数包裹批次训练过程，形成轮次训练过程

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized or loaded from
            the existed model
        begin_epoch: int
            The begin epoch of this train procession
        end_epoch: int
            The end epoch of this train procession
        batch_size: int
            The size of each batch
        train_data: Iterable
            The data used for this train procession,
            NOTICE: should have been divided to batches
        trainer:
            The trainer used to update the parameters of the net
        bp_loss_f: dict with only one value and one key
            The function to compute the loss for the procession
            of back propagation
        loss_function: dict of function
            Some other measurement in addition to bp_loss_f
        test_data: Iterable
            The data used for the evaluation at the end of each epoch,
            NOTICE: should have been divided to batches
            Default to ``None``
        ctx: Context or list of Context
            Defaults to ``mx.cpu()``.
        toolbox: Toolbox
            Default to ``None``
        kwargs
        """
        # 参数修改时需要同步修改 fit 函数中的参数
        # 定义轮次训练过程
        if toolbox is not None:
            epoch_timer = toolbox.get('timer')
            formatter = toolbox.get('formatter')
        else:
            epoch_timer = None
            formatter = None

        for epoch in range(begin_epoch, end_epoch):
            if epoch_timer:
                epoch_timer.start()

            loss_values = self.batch_loop(
                net=net, epoch=epoch, batch_size=batch_size,
                train_data=train_data,
                trainer=trainer, bp_loss_f=bp_loss_f,
                loss_function=loss_function,
                ctx=ctx,
                toolbox=toolbox,
            )

            train_time = epoch_timer.end(wall=True) if epoch_timer else None

            # # todo 定义每一轮结束后的模型评估方法
            evaluation_result = Module.eval(
                net, test_data, ctx=ctx
            )
            if formatter:
                evaluation_formatter = formatter.get('evaluation')
                if evaluation_formatter:
                    print(
                        evaluation_formatter(
                            epoch=epoch,
                            train_time=train_time,
                            loss_name_value=loss_values,
                            eval_name_value=evaluation_result,
                            extra_info=None,
                            dump=True,
                        )[0]
                    )

            # todo 定义模型保存方案
            if kwargs.get('prefix') and (
                    epoch % kwargs.get('save_epoch', 1) == 0 or
                    end_epoch - 10 <= epoch <= end_epoch - 1
            ):
                self.save_params(
                    kwargs['prefix'] + "-%04d.parmas" % (epoch + 1), net
                )

    def batch_loop(
            self,
            net, epoch, batch_size,
            train_data,
            trainer, bp_loss_f,
            loss_function,
            ctx=mx.cpu(),
            toolbox=None,
    ):
        """
        The true body of batch loop

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized
            or loaded from the existed model
        epoch: int
            Current Epoch
        batch_size: int
            The size of each batch
        train_data: Iterable
            The data used for this train procession,
            NOTICE: should have been divided to batches
        trainer:
            The trainer used to update the parameters of the net
        bp_loss_f: dict with only one value and one key
            The function to compute the loss for the procession
            of back propagation
        loss_function: dict of function
            Some other measurement in addition to bp_loss_f
        ctx: Context or list of Context
            Defaults to ``mx.cpu()``.
        toolbox: dict
            Default to ``None``

        Returns
        -------
        loss_values: dict
            Recorded the loss values computed by the loss_function
        """
        loss_monitor = None
        progress_monitor = None

        if toolbox is not None:
            monitor = toolbox.get("monitor")
            if monitor is not None:
                loss_monitor = monitor.get("loss")
                progress_monitor = monitor.get("progress")

        if progress_monitor is not None:
            progress_monitor.batch_start(epoch)

        for i, batch_data in enumerate(train_data):
            self.fit_f(
                net=net, batch_size=batch_size, batch_data=batch_data,
                trainer=trainer, bp_loss_f=bp_loss_f,
                loss_function=loss_function,
                loss_monitor=loss_monitor,
                ctx=ctx,
            )
            if progress_monitor is not None:
                loss_values = [loss for loss in loss_monitor.values()]
                progress_monitor(i, loss_value=loss_values)
        progress_monitor.batch_end()
        loss_values = {
            name: loss for name, loss in loss_monitor.items()
        }
        return loss_values

    @staticmethod
    def eval(net, test_data, ctx=mx.cpu()):
        """
        在这里定义数据评估方法

        Parameters
        ----------
        net
        test_data
        ctx

        Returns
        -------
        metrics: dict

        """
        return eval_f(net, test_data, ctx)

    @staticmethod
    def fit_f(net, batch_size, batch_data,
              trainer, bp_loss_f, loss_function, loss_monitor=None,
              ctx=mx.cpu()
              ):
        """
        Defined how each step of batch train goes

        Parameters
        ----------
        net: HybridBlock
            The network which has been initialized
            or loaded from the existed model
        batch_size: int
                The size of each batch
        batch_data: Iterable
            The batch data for train
        trainer:
            The trainer used to update the parameters of the net
        bp_loss_f: dict with only one value and one key
            The function to compute the loss for the procession
            of back propagation
        loss_function: dict of function
            Some other measurement in addition to bp_loss_f
        loss_monitor: LossMonitor
            Default to ``None``
        ctx: Context or list of Context
            Defaults to ``mx.cpu()``.

        Returns
        -------

        """
        # 此函数定义训练过程
        ctx_data = Module.split_and_load(
            ctx, *batch_data,
            even_split=False
        )

        bp_loss = None
        with autograd.record():
            # todo modify the component extracted from ctx_data
            for _data in ctx_data:
                bp_loss = fit_f(
                    net, _data, bp_loss_f, loss_function, loss_monitor
                )
            assert bp_loss is not None
            bp_loss.backward()
        trainer.step(batch_size)

    @staticmethod
    def split_and_load(ctx, *args, **kwargs):
        return split_and_load(ctx, *args, **kwargs)



# coding: utf-8
# create by tongshiwei on 2019/4/12

from mxnet import gluon
from tqdm import tqdm

__all__ = ["load_data", "transform", "get_data_iter"]


def load_data(data_src):
    raise NotImplementedError


def transform(raw_data, params):
    # 定义数据转换接口
    # raw_data --> batch_data

    batch_size = params.batch_size

    return gluon.data.DataLoader(gluon.data.ArrayDataset(raw_data), batch_size)


def get_data_iter(*args, **kwargs):
    raw_data = load_data(*args)
    transform(raw_data, **kwargs)
    raise NotImplementedError


if __name__ == '__main__':
    for data in tqdm(get_data_iter()):
        pass



# coding: utf-8
# created by tongshiwei on 18-2-5
import math
from contextlib import contextmanager

from longling.lib.stream import flush_print

try:
    NAN = math.nan
except (AttributeError, ImportError):
    NAN = float('nan')

__all__ = ["ConsoleProgressMonitor"]


class ConsoleProgressMonitor(object):
    def __init__(self, loss_index=None, eval_index=None, batch_num=NAN,
                 end_epoch=NAN, output=True, silent=False,
                 **kwargs):
        loss_index = [] if loss_index is None else loss_index
        eval_index = [] if eval_index is None else eval_index

        self.batch_num = batch_num
        self._batch_num = None
        self.end_epoch = end_epoch

        self.eval_index = eval_index
        self.loss_index = ["Loss-%s" % ln for ln in loss_index]

        self.output = output

        info_header = "{:>5}| {:>7}" + " " * 5 \
                      + "{:>10}" + " " * 2 + "{:>10}" + " " * 5
        loss_header = (" " * 2).join(
            ["{:>%s}" % min(len(index), 15) for index in loss_index]
        )
        eval_header = (" " * 2).join(
            ["{:>%s}" % min(len(index), 15) for index in eval_index]
        )

        self.output_formatter = info_header + loss_header + eval_header

        arguments = list(self.loss_index) + list(self.eval_index)
        self.index_prefix = self.output_formatter.format("Epoch", "Total-E",
                                                         "Batch", "Total-B",
                                                         *arguments)

        self.epoch = NAN
        self.silent = silent

    def __call__(self, batch_no, loss_value=None, eval_value=None):
        loss_value = [] if loss_value is None else loss_value
        eval_value = [] if eval_value is None else eval_value

        arguments = list(loss_value) + list(eval_value)
        res_str = self.output_formatter.format(self.epoch, self.end_epoch,
                                               batch_no, self.batch_num,
                                               *arguments)

        if self.output and not self.silent:
            flush_print(res_str)

        self._batch_num = batch_no if self._batch_num is None else max(
            [batch_no, self._batch_num]
        )
        return res_str

    def batch_start(self, epoch):
        res_str = self.index_prefix
        self.epoch = epoch

        if self.output and not self.silent:
            print(res_str)

        return res_str

    def batch_end(self, batch_num=None):
        if batch_num is not None:
            self.batch_num = batch_num
        elif self._batch_num is not None:
            self.batch_num = self._batch_num
        if self.output and not self.silent:
            print("")

        return ""

    @contextmanager
    def watching(self, epoch):
        self.batch_start(epoch)
        yield
        self.batch_end()
